# Copyright 2021 The KCL Authors. All rights reserved.

from typing import cast

import kclvm.api.object as obj


def value_subsume(
    value1: obj.KCLObject, value2: obj.KCLObject, should_recursive_check: bool = True
) -> bool:
    """Calculate the partial order relationship between `KCL value objects`
    and judge whether the value1 object ∈ the value2 object.

    Please note that The type and value of KCL are defined and used separately,
    so the partial order relationship calculation is also divided into two types,
    type and value, and there is no partial order relationship between type
    objects and value objects.
    """
    if not value1 or not value2:
        return False
    if not isinstance(value1, obj.KCLObject) or not isinstance(value2, obj.KCLObject):
        return False
    if value1 == value2 or value1 is value2:
        return True
    if isinstance(value1, (obj.KCLNoneObject, obj.KCLUndefinedObject)):
        return True
    if isinstance(value2, (obj.KCLNoneObject, obj.KCLUndefinedObject)):
        return True
    if isinstance(value1, obj.KCLIntObject):
        return isinstance(value2, obj.KCLIntObject) and value1.value == value2.value
    if isinstance(value1, obj.KCLFloatObject):
        return isinstance(value2, obj.KCLFloatObject) and value1.value == value2.value
    if isinstance(value1, obj.KCLNameConstantObject):
        return (
            isinstance(value2, obj.KCLNameConstantObject)
            and value1.value == value2.value
        )
    if isinstance(value1, (obj.KCLListObject, obj.KCLTupleObject)):
        return (
            isinstance(value2, (obj.KCLListObject, obj.KCLTupleObject))
            and len(value1.items) == len(value2.items)
            and all(
                [
                    value_subsume(item1, item2, should_recursive_check)
                    for item1, item2 in zip(value1.items, value2.items)
                ]
            )
        )
    if isinstance(value1, (obj.KCLDictObject, obj.KCLSchemaObject)):
        if isinstance(value2, (obj.KCLDictObject, obj.KCLSchemaObject)):
            value1_dict = {k: value1.get(k) for k in sorted(list(value1.value.keys()))}
            value2_dict = {k: value2.get(k) for k in sorted(list(value2.value.keys()))}

            if len(value1_dict) == 0:
                return True

            if all([key not in value2_dict for key in value1_dict]):
                return True

            if should_recursive_check:
                for key1, value1 in value1_dict.items():
                    if key1 not in value2_dict:
                        continue
                    value2 = value2_dict.get(key1)
                    if not value_subsume(value1, value2, should_recursive_check):
                        return False
            return True
    return False


def type_subsume(
    value1: obj.KCLObject, value2: obj.KCLObject, check_left_any: bool = False
) -> bool:
    """Calculate the partial order relationship between `KCL type objects`
    and judge whether the value1 object ∈ the value2 object.

    Please note that The type and value of KCL are defined and used separately,
    so the partial order relationship calculation is also divided into two types,
    type and value, and there is no partial order relationship between type
    objects and value objects.
    """
    if not value1 or not value2:
        return False
    if not isinstance(value1, obj.KCLObject) or not isinstance(value2, obj.KCLObject):
        return False
    if value1 == value2 or value1 is value2:
        return True
    if check_left_any and isinstance(value1, obj.KCLAnyTypeObject):
        return True
    if isinstance(value2, obj.KCLAnyTypeObject):
        return True
    if isinstance(value1, obj.KCLNoneTypeObject):
        return True
    if isinstance(value1, obj.KCLUnionTypeObject):
        return all([type_subsume(tpe, value2) for tpe in value1.types])
    if isinstance(value2, obj.KCLUnionTypeObject):
        return any([type_subsume(value1, tpe) for tpe in value2.types])
    if isinstance(value1, obj.KCLSchemaTypeObject):
        if not isinstance(value2, obj.KCLSchemaTypeObject):
            return False
        value1 = cast(obj.KCLSchemaTypeObject, value1)
        value2 = cast(obj.KCLSchemaTypeObject, value2)
        return value1.is_sub_schema_of(value2)
    if isinstance(value1, obj.KCLIntTypeObject) and isinstance(
        value2, obj.KCLFloatTypeObject
    ):
        return True
    if isinstance(value1, obj.KCLBuiltinTypeObject):
        return (
            isinstance(value2, obj.KCLBuiltinTypeObject)
            and value1.type_kind() == value2.type_kind()
        )
    if isinstance(
        value1,
        (
            obj.KCLStringLitTypeObject,
            obj.KCLNumberLitTypeObject,
            obj.KCLBoolLitTypeObject,
        ),
    ):
        if isinstance(
            value2,
            (
                obj.KCLStringLitTypeObject,
                obj.KCLNumberLitTypeObject,
                obj.KCLBoolLitTypeObject,
            ),
        ):
            return (
                value1.type_kind() == value2.type_kind()
                and value1.value == value2.value
            )
        elif isinstance(value2, obj.KCLBuiltinTypeObject):
            # float_lit -> float
            # int_lit -> int
            # bool_lit -> bool
            # str_lit -> str
            # int_lit/bool_lit -> float
            if isinstance(value2, obj.KCLFloatTypeObject) and not isinstance(
                value1, obj.KCLStringLitTypeObject
            ):
                return True
            return value2.type_str() in value1.type_str()
    if isinstance(value1, obj.KCLListTypeObject):
        return isinstance(value2, obj.KCLListTypeObject) and type_subsume(
            value1.item_type, value2.item_type, check_left_any
        )
    if isinstance(value1, obj.KCLDictTypeObject):
        return (
            isinstance(value2, obj.KCLDictTypeObject)
            and type_subsume(value1.key_type, value2.key_type, check_left_any)
            and type_subsume(value1.value_type, value2.value_type, check_left_any)
        )
    if isinstance(value1, obj.KCLNumberMultiplierTypeObject) and isinstance(
        value2, obj.KCLNumberMultiplierTypeObject
    ):
        if value1.is_literal():
            return not value2.is_literal() and value1.type_str() == value2.type_str()
        else:
            return True
    return False



# Copyright 2022 The KCL Authors. All rights reserved.

import os
import sys
import platform
import typing
import json
import inspect

from ctypes import *

import kclvm.compiler.extension.plugin.plugin as kcl_plugin
import kclvm.kcl.info as kcl_info
import kcl_lib.api.spec_pb2 as pb2
import kclvm.api.object as objpkg
import kcl_lib.api as api
from ruamel.yaml import YAML

# Ruamel YAML instance
ruamel_yaml = YAML(typ="unsafe", pure=True)
# Convert None to null
ruamel_yaml.representer.add_representer(
    type(None),
    lambda dumper, data: dumper.represent_scalar("tag:yaml.org,2002:null", "null"),
)

# Create the api instance
_api = api.API()

kclvm_PANIC_INFO_KEY = "__kcl_PanicInfo__"

# Using kclvm rust cli PATH or current exec path.
_exe_root = os.path.dirname(
    os.environ.get("KCLVM_CLI_BIN_PATH") or os.path.dirname(sys.executable)
)
_cli_dll = None


def init_cli_dll():
    global _cli_dll

    if _cli_dll:
        return

    if platform.system() == "Darwin":
        if os.path.exists(f"{_exe_root}/lib/libkclvm_cli_cdylib.dylib"):
            _cli_dll_path = f"{_exe_root}/lib/libkclvm_cli_cdylib.dylib"
        else:
            _cli_dll_path = f"{_exe_root}/bin/libkclvm_cli_cdylib.dylib"
    elif platform.system() == "Linux":
        if os.path.exists(f"{_exe_root}/lib/libkclvm_cli_cdylib.so"):
            _cli_dll_path = f"{_exe_root}/lib/libkclvm_cli_cdylib.so"
        else:
            _cli_dll_path = f"{_exe_root}/bin/libkclvm_cli_cdylib.so"
    elif platform.system() == "Windows":
        _cli_dll_path = f"{_exe_root}\\bin\\kclvm_cli_cdylib.dll"
    else:
        raise f"unknown os: {platform.system()}"
    _cli_dll = CDLL(_cli_dll_path)


class PluginContext:
    def __init__(self):
        self._plugin_dict: typing.Dict[str, any] = {}

    def call_method(self, name: str, args_json: str, kwargs_json: str) -> str:
        return self._call_py_method(name, args_json, kwargs_json)

    def _call_py_method(self, name: str, args_json: str, kwargs_json: str) -> str:
        try:
            return self._call_py_method_unsafe(name, args_json, kwargs_json)
        except Exception as e:
            return json.dumps({"__kcl_PanicInfo__": f"{e}"})

    def _get_plugin(self, plugin_name: str) -> typing.Optional[any]:
        if plugin_name in self._plugin_dict:
            return self._plugin_dict[plugin_name]

        module = kcl_plugin.get_plugin(plugin_name)
        self._plugin_dict[plugin_name] = module
        return module

    def _call_py_method_unsafe(
        self, name: str, args_json: str, kwargs_json: str
    ) -> str:
        dotIdx = name.rfind(".")
        if dotIdx < 0:
            return ""

        modulePath = name[:dotIdx]
        mathodName = name[dotIdx + 1 :]

        plugin_name = modulePath[modulePath.rfind(".") + 1 :]

        module = self._get_plugin(plugin_name)
        mathodFunc = None

        for func_name, func in inspect.getmembers(module):
            if func_name == kcl_info.demangle(mathodName):
                mathodFunc = func
                break

        args = []
        kwargs = {}

        if args_json:
            args = json.loads(args_json)
            if not isinstance(args, list):
                return ""
        if kwargs_json:
            kwargs = json.loads(kwargs_json)
            if not isinstance(kwargs, dict):
                return ""

        result = mathodFunc(*args, **kwargs)
        return json.dumps(result)


__plugin_context__ = PluginContext()
__plugin_method_agent_buffer__ = create_string_buffer(1024)


@CFUNCTYPE(c_char_p, c_char_p, c_char_p, c_char_p)
def plugin_method_agent(method: str, args_json: str, kwargs_json: str) -> c_char_p:
    method = str(method, encoding="utf-8")
    args_json = str(args_json, encoding="utf-8")
    kwargs_json = str(kwargs_json, encoding="utf-8")

    json_result = __plugin_context__.call_method(method, args_json, kwargs_json)

    global __plugin_method_agent_buffer__
    __plugin_method_agent_buffer__ = create_string_buffer(json_result.encode("utf-8"))
    return addressof(__plugin_method_agent_buffer__)


def kclvm_cli_native_run_dylib(args: pb2.ExecProgram_Args) -> objpkg.KCLResult:
    result: pb2.ExecProgram_Result = _api.exec_program(args)
    if result.err_message:
        raise Exception(result.err_message)

    if result.log_message:
        print(result.log_message)

    data = list(ruamel_yaml.load_all(result.yaml_result))

    return objpkg.KCLResult(data, os.path.abspath(args.k_filename_list[-1]))



# Copyright 2021 The KCL Authors. All rights reserved.

import re
import pathlib

from io import StringIO
from typing import List, Union

import kclvm.kcl.error as kcl_error
import kclvm.internal.log as klog

from kclvm.compiler.parser.lark_tree import get_lark_tree_from_expr
from kclvm.compiler.parser.lark_tree import (
    Token,
    Tree,
    AstType,
    TreeWalker,
    TOKEN_TYPE,
    TREE_TYPE,
    OPERATOR_TOKENS,
)

_INVALID_NEWLINE_STRING_MSG = "invalid NEWLINE token string value {}"
_INLINE_COMMENT_REGEX = "#[^\n]*\n"
_INLINE_COMMENT_WITH_MULTILINE_REGEX = "#[^\n]*[\n\t ]*"
_NODE_WITH_NEWLINE_EXPRS = [
    Tree.CONFIG_EXPR,
    Tree.DICT_COMP,
    Tree.DICT_EXPR,
    Tree.LIST_COMP,
    Tree.LIST_EXPR,
    Tree.COMP_CLAUSE,
    Tree.SIMPLE_STMT,
    Tree.SCHEMA_MEMBER_STMT,
    Tree.MIXINS,
]
SEPARATOR_TOKEN = " "
EMPTY_TOKEN = ""
ENDLINE_TOKEN = "\n"
COMMENT_START_TOKEN = "#"


class TextAdapterWalker(TreeWalker):
    """
    Walker adapted for text processing, can be used as a semantic-independent
    walker super class such as formatter and linter
    """

    def __init__(self):
        super().__init__()
        self.printer = StringIO()  # Printer used to write expressions and tokens
        self.indent_level: int = 0  # Now indent level
        self.indent_queue = [0]  # Indent queue
        self.indent_space_count: int = 4  # Default indent space count

    def write(self, text: str) -> None:
        """Append a piece of text to the current line."""
        self.printer.write(text)

    def write_token_separator(self) -> None:
        """Print the separator between tokens."""
        self.write(SEPARATOR_TOKEN)

    def walk_node(self, node: Union[AstType, str]) -> None:
        """Write node"""
        if isinstance(node, str):
            self.write(node)
        else:
            self.walk(node)

    def fill(self, text: str = "") -> None:
        """Append a piece of text to the current line."""
        self.printer.write(ENDLINE_TOKEN)
        self.write_indent()
        self.printer.write(text)

    def write_indent(self) -> None:
        """Append indent white space according indent level"""
        self.printer.write(
            self.indent_space_count * self.indent_level * SEPARATOR_TOKEN
        )

    def count_blank_line(self, text: str) -> int:
        """Blank line count in a NEWLINE token"""
        return re.sub(_INLINE_COMMENT_REGEX, EMPTY_TOKEN, text).count(ENDLINE_TOKEN) - 1

    def count_indent(self, text: str) -> int:
        """
        Count the indent by number of white spaces for a leading text
        e.g. NEWLINE "\n    " its indent count is 4
        e.g. NEWLINE "\n # inline comment\n   " its indent count is 3
        """
        if ENDLINE_TOKEN not in text:
            return 0
        line = text.split(ENDLINE_TOKEN)[-1]
        temptext = line.replace(SEPARATOR_TOKEN, EMPTY_TOKEN)
        count = len(line) if len(temptext) == 0 else 0
        if count not in self.indent_queue:
            # Indent increase
            self.indent_queue.append(count)
        else:
            # Indent decrease
            idx = self.indent_queue.index(count)
            del self.indent_queue[idx + 1 :]
        idx = self.indent_queue.index(count)
        self.indent_level = idx
        return count

    def generic_walk(self, node: AstType) -> None:
        """Called if no explicit walker function exists for a node."""
        if node["type"] == TOKEN_TYPE:
            self.walk_pre_token(node)
            self.walk_token(node)
            self.walk_post_token(node)
        elif node["type"] == TREE_TYPE:
            for n in node["children"]:
                self.walk(n)
        else:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CompileError_TYPE,
                arg_msg="Unknown format node type",
            )

    def walk_pre_token(self, node: AstType) -> None:
        """Deal after token"""
        pass

    def walk_post_token(self, node: AstType) -> None:
        """Deal after token"""
        pass

    def walk_token(self, node: AstType) -> None:
        """AST: token"""
        pass


class Formatter(TextAdapterWalker):
    def __init__(self) -> None:
        super().__init__()
        self.last_token: str = EMPTY_TOKEN  # Last token name
        self.last_token_value: str = EMPTY_TOKEN  # Last token value
        self.single_comment_spaces: int = 2  # Inline comment after its expressions
        self.max_blank_line: int = 1  # Maximum number of blank lines
        self._is_in_arguments = False  # Mark is in arguments
        self._is_in_collection_if = False  # Mark is in collection if
        self._is_colon_without_space = False  # Mark is in expr with colon except space

    # Base

    def generic_walk(self, node: AstType) -> None:
        """Formatter special walk"""
        if node["name"] in _NODE_WITH_NEWLINE_EXPRS:
            self.write_node_with_newline(node)
        else:
            super().generic_walk(node)

    # Tokens

    def add_blank_line(self) -> None:
        """Blank line before and after schema_stmt"""
        if (
            self.last_token == Token.NEWLINE
            and self.count_blank_line(self.last_token_value) < self.max_blank_line
        ):
            self.write_token(ENDLINE_TOKEN)
            self.last_token = Token.NEWLINE
            self.last_token_value = ENDLINE_TOKEN * 2

    def split_newline_value(self, value: str) -> List[str]:
        """Split a NEWLINE token value into newline parts and inline comment parts

        Input: "\n \n # comment \n # comment \n\n # comment \n"
        Output: ["\n \n ", "# comment ", "\n ", "# comment ", "\n\n ", "# comment ", "\n"]
        """
        if not value:
            return []
        parts = []
        # Mark containing COMMENT token
        index = value.find(COMMENT_START_TOKEN)
        if index == -1:
            return [value]  # Single NEWLINE without COMMENT
        elif index > 0:
            parts.append(value[:index])  # Add first NEWLINE token
        inline_comments = re.findall(_INLINE_COMMENT_WITH_MULTILINE_REGEX, value)
        for comment in inline_comments:
            index = comment.find(ENDLINE_TOKEN)
            if index == -1:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.CompileError_TYPE,
                    arg_msg=_INVALID_NEWLINE_STRING_MSG.format(comment),
                )
            parts.append(comment[:index])  # Add COMMENT token
            parts.append(comment[index:])  # Add NEWLINE token
        if len(parts) > 1 and ENDLINE_TOKEN not in parts[-1]:
            # Last part is not NEWLINE, raise an error
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CompileError_TYPE,
                arg_msg=_INVALID_NEWLINE_STRING_MSG.format(value),
            )
        return parts

    def write_newline(self, value: str) -> None:
        """
        Write newline except inline comment
        """
        blank_line_count = self.count_blank_line(value)
        value = value.replace(SEPARATOR_TOKEN, "")
        newline_count = 0
        for i, c in enumerate(value):
            if c == ENDLINE_TOKEN:
                newline_count += 1
                # Consecutive blank lines
                if newline_count < blank_line_count:
                    pass
                # If there is a blank line, keep it
                elif self.count_blank_line(value[i + 1 :]):
                    self.fill()
                # First NEWLINE in last expr
                else:
                    self.write(ENDLINE_TOKEN)

    def walk_newline(self, token: AstType) -> None:
        """
        Format NEWLINE token contains '#', '\n' and ' '
        """
        # Record indent level and Windows line break handling and 1 tab <-> 4 spaces
        value = token["value"].replace("\r", ENDLINE_TOKEN).replace("\t", "    ")
        # Remove start blank lines except comments and first token is NEWLINE
        if self.last_token == "":
            # Get all inline comments
            self.write("".join(re.findall(_INLINE_COMMENT_REGEX, value)))
            return
        # Record indent
        self.count_indent(value)
        parts = self.split_newline_value(value)
        for part in parts:
            if part.startswith(COMMENT_START_TOKEN):
                self.write(part)
            else:
                self.write_newline(part)

    def write_string(self, value: str) -> None:
        """Print KCL string with prefix"""
        quota = False
        for c in value:
            # The string prefix is uniformly lowercase
            if not quota:
                c = c.lower()
            # String start quotation marks
            if c == '"' or c == "'":
                quota = True
            self.write(c)

    def write_token(self, node: Union[str, AstType]) -> None:
        """Write token node or token string"""
        if not node:
            return
        if isinstance(node, str):
            self.write(node)
        else:
            self.walk_token(node)

    def walk_post_token(self, token: AstType) -> None:
        """Deal after token"""
        name, value = token["name"], token["value"]
        if name == Token.ASSIGN and self._is_in_arguments:
            # Do not write space between '=' in kwargs
            pass
        elif (
            name == Token.COLON
            and self._is_colon_without_space
            and not self._is_in_arguments
        ):
            # Do not write space after colon ':' e.g. check: and schema Name:
            pass
        elif (
            name in [Token.PLUS, Token.MINUS, Token.NOT]
            and self.last_token in OPERATOR_TOKENS
        ):
            # Do not write space between unary operator
            pass
        elif name in OPERATOR_TOKENS:
            self.write_token_separator()
        elif name == Token.FOR and self.last_token != Token.NEWLINE:
            self.write_token_separator()
        elif name in [Token.IF, Token.ELIF] and self._is_in_collection_if:
            self.write_token_separator()
        # Write space after , and : except NEWLINE behind it and write space behind keyword
        elif name in [
            Token.IMPORT,
            Token.LAMBDA,
            Token.SCHEMA,
            Token.RELAXED,
            Token.ASSERT,
            Token.FINAL,
            Token.IN,
            Token.AS,
            Token.IS,
            Token.MIXIN,
            Token.PROTOCOL,
            Token.TYPE,
            Token.COMMA,
            Token.COLON,
            Token.L_AND,
            Token.L_OR,
            Token.L_L_NOT,
            Token.SEMI_COLON,
            Token.ALL,
            Token.ANY,
            Token.MAP,
            Token.FILTER,
        ]:
            self.write_token_separator()
        self.last_token, self.last_token_value = name, value

    def walk_pre_token(self, token: AstType) -> None:
        """Deal before token"""
        name = token["name"]
        value = token["value"]
        if name == Token.ASSIGN and self._is_in_arguments:
            # Do not write space between '=' in kwargs
            pass
        elif (
            name in [Token.PLUS, Token.MINUS, Token.NOT]
            and self.last_token in OPERATOR_TOKENS
        ):
            # Do not write space between unary operator
            pass
        elif name in OPERATOR_TOKENS:
            self.write_token_separator()
        elif name == Token.FOR and self.last_token != Token.NEWLINE:
            self.write_token_separator()
        elif name in [Token.IN, Token.AS, Token.IS, Token.L_AND, Token.L_OR]:
            self.write_token_separator()
        elif (
            name == Token.NEWLINE
            and value.startswith(COMMENT_START_TOKEN)
            and self.last_token != ""
        ):
            # Two spaces between expr and inline comment #
            self.write(SEPARATOR_TOKEN * self.single_comment_spaces)

    def walk_token(self, token: AstType) -> None:
        """AST: token node"""
        name = token["name"]
        value = token["value"]
        if name == Token.NEWLINE:
            self.walk_newline(token)
        elif Token.is_string(name):
            self.write_string(value)
        else:
            self.write_token(value)

    # Exprs

    def write_expr(self, node: AstType) -> None:
        """Write expr"""
        if node:
            self.walk_nodes(*node["children"])

    def write_arguments(self, nodes: List[AstType]) -> None:
        """Arguments in call_expr, str_expr and schema_expr"""
        self._is_in_arguments = True
        self.walk_nodes(*nodes)
        self._is_in_arguments = False

    def walk_call_suffix(self, node: AstType) -> None:
        """AST: call_suffix

        call_suffix: LEFT_PARENTHESES [arguments [COMMA]] RIGHT_PARENTHESES
        """
        self.write_arguments(node["children"])

    def walk_schema_expr(self, node: AstType) -> None:
        """AST: schema_expr

        schema_expr: identifier (LEFT_PARENTHESES [arguments] RIGHT_PARENTHESES)? dict_expr
        """
        self.write_expr(self.get(node, Tree.IDENTIFIER))
        self.write_arguments(
            self.get_internal(node, Token.LEFT_PARENTHESES, Token.RIGHT_PARENTHESES)
        )
        # Write space between 'SchemaName' and '{}' in schema_expr
        self.write_token_separator()
        self.write_expr(self.get(node, Tree.CONFIG_EXPR))

    def walk_star_expr(self, node: AstType) -> None:
        """AST: star_expr

        star_expr: MULTIPLY test
        """
        # Do not write space between * in list iter and var
        if self.get(node, Token.MULTIPLY):
            self.write("*")
        self.write_expr(self.get(node, Tree.TEST))

    def walk_double_star_expr(self, node: AstType) -> None:
        """AST: double_star_expr

        double_star_expr: DOUBLE_STAR test
        """
        # Do not write space between ** in dict iter and var
        if self.get(node, Token.DOUBLE_STAR):
            self.write("**")
        self.write_expr(self.get(node, Tree.TEST))

    def walk_lambda_expr(self, node: AstType) -> None:
        """AST: if_expr

        lambda_expr: LAMBDA [schema_arguments] [RIGHT_ARROW type] LEFT_BRACE [expr_stmt | NEWLINE _INDENT schema_init_stmt+ _DEDENT] RIGHT_BRACE
        """
        for n in node["children"]:
            self.walk_node(n)
            if n["name"] in [Tree.SCHEMA_ARGUMENTS, Token.RIGHT_ARROW, Tree.TYPE]:
                self.write_token_separator()

    def walk_if_expr(self, node: AstType) -> None:
        """AST: if_expr

        if_expr: or_test IF or_test ELSE test
        """
        self.write_expr(self.get(node, Tree.SIMPLE_EXPR))
        while self.get(node, Token.IF):
            self.write_token_separator()
            self.write_token("if")
            self.write_token_separator()
            self.write_expr(self.get(node, Tree.SIMPLE_EXPR))
            self.write_token_separator()
            self.write_token("else")
            self.write_token_separator()
            self.write_expr(self.get(node, Tree.TEST))

    def walk_comp_clause(self, node: AstType) -> None:
        """AST: comp_clause

        comp_clause: FOR loop_variables [COMMA] IN or_test [NEWLINE] [IF test [NEWLINE]]
        """
        for n in node["children"]:
            if n["name"] == Token.IF and self.last_token != Token.NEWLINE:
                self.write_token_separator()
            self.walk_node(n)
            if n["name"] == Token.IF:
                self.write_token_separator()

    def walk_quant_expr(self, node: AstType) -> None:
        """AST: quant_expr

        quant_expr: quant_op [ identifier COMMA ] identifier IN quant_target LEFT_BRACE (simple_expr [IF simple_expr] | NEWLINE _INDENT simple_expr [IF simple_expr] NEWLINE _DEDENT)? RIGHT_BRACE
        """
        for n in node["children"]:
            if n["name"] == Token.IF and self.last_token != Token.NEWLINE:
                self.write_token_separator()
            if n["name"] == Token.LEFT_BRACE:
                self.write_token_separator()
            self.walk_node(n)
            if n["name"] == Token.IF:
                self.write_token_separator()

    def write_node_with_newline(self, node: AstType) -> None:
        """
        Write node with , ; and NEWLINE
        such as list_comp and dict_comp.
        """
        children = node["children"]
        for i, n in enumerate(children):
            if (
                n["name"] in [Token.COMMA, Token.SEMI_COLON]
                and i < len(children) - 1
                and children[i + 1]["name"] == Token.NEWLINE
            ):
                self.write(n["value"])
            else:
                self.walk(n)

    def write_colon_without_space(self, node: AstType) -> None:
        """
        Write expr with colon : without space
        """
        self._is_colon_without_space = True
        self.write_expr(node)
        self._is_colon_without_space = False

    def walk_mixins(self, node: AstType) -> None:
        """AST: mixins

        mixins: identifier (COMMA (NEWLINE mixins | identifier))*
        """
        self.write_node_with_newline(node)

    def walk_list_items(self, node: AstType) -> None:
        """AST: list_items

        list_items: list_item ((COMMA [NEWLINE] | NEWLINE) list_item)* [COMMA] [NEWLINE]
        """
        self.write_node_with_newline(node)

    def walk_entries(self, node: AstType) -> None:
        """AST: entries

        entries: entry ((COMMA [NEWLINE] | NEWLINE) entry)* [COMMA] [NEWLINE]
        """
        self.write_node_with_newline(node)

    def walk_config_entries(self, node: AstType) -> None:
        """AST: config_entries

        entries: entry ((COMMA [NEWLINE] | NEWLINE) entry)* [COMMA] [NEWLINE]
        """
        self.write_node_with_newline(node)

    def walk_subscript_suffix(self, node: AstType) -> None:
        """AST: subscript_suffix

        subscript_suffix: LEFT_BRACKETS (test | [test] COLON [test] [COLON [test]]) RIGHT_BRACKETS
        """
        self.write_colon_without_space(node)

    def walk_slice_suffix(self, node: AstType) -> None:
        """AST: slice_suffix

        slice_suffix: LEFT_BRACKETS (test | [test] COLON [test] [COLON [test]]) RIGHT_BRACKETS
        """
        self.write_colon_without_space(node)

    def walk_bin_op(self, node: AstType) -> None:
        """AST: comp_op"""
        children = node["children"]
        for i, n in enumerate(children):
            # IN | L_NOT IN | IS | IS L_NOT | L_NOT
            if n["name"] == Token.IS:
                self.write_token_separator()
                self.write(n["value"])
                if i + 1 < len(children) and children[i + 1]["name"] != Token.L_L_NOT:
                    self.write_token_separator()
            elif n["name"] == Token.L_L_NOT:
                self.write_token_separator()
                if i == 0:
                    self.write(n["value"])
                    if i + 1 < len(children) and children[i + 1]["name"] != Token.IN:
                        self.write_token_separator()
                if i == 1:
                    self.write("not")
                    self.write_token_separator()
            else:
                self.walk(n)

    def walk_not_test(self, node: AstType) -> None:
        """AST: not_test"""
        if self.get(node, Token.L_NOT):
            self.write_token("not")
            self.write_token_separator()
        self.write_expr(node)

    def walk_check_block(self, node: AstType) -> None:
        """AST: check_block

        check_block: CHECK COLON NEWLINE _INDENT check_expr+ _DEDENT
        """
        self.write_token(self.get(node, Token.CHECK))
        self.write_token(self.get(node, Token.COLON))
        self.write_expr(node)

    def walk_check_expr(self, node: AstType) -> None:
        """AST: check_expr

        check_expr: simple_expr [IF simple_expr] [COMMA primary_expr] NEWLINE
        """
        self.write_expr(self.get(node, Tree.SIMPLE_EXPR))
        if self.get(node, Token.IF):
            self.write_token_separator()
            self.write_token("if")
            self.write_token_separator()
            self.write_expr(self.get(node, Tree.SIMPLE_EXPR))
        for n in node["children"]:
            self.walk(n)

    def walk_dict_type(self, node: AstType) -> None:
        """AST: dict_type"""
        self.write_colon_without_space(node)

    # Statements

    def write_stmt(self, node: AstType) -> None:
        """Write stmt"""
        self.write_token(self.get(node, Token.NEWLINE))
        self.walk_nodes(*node["children"])

    def write_condition_and_body(self, node: AstType) -> None:
        """Write if and elif condition and body"""
        self.write_token_separator()
        self.write_expr(self.get(node, Tree.TEST))
        self.write_token(":")
        self.write_stmt(self.get(node, Tree.EXECUTION_BLOCK))

    def walk_start(self, node: AstType) -> None:
        """node: start
        start: (NEWLINE | statement)*
        statement: simple_stmt | compound_stmt
        simple_stmt: (assign_stmt | unification_stmt | expr_stmt | assert_stmt | import_stmt | type_alias_stmt) NEWLINE
        """
        last_stmt_is_import = False
        for n in node["children"]:
            if n["name"] != Token.NEWLINE:
                stmt_name = n["children"][0]["children"][0]["name"]
                # Add a blank line after consecutive import statements
                if last_stmt_is_import and stmt_name != Tree.IMPORT_STMT:
                    self.add_blank_line()
                last_stmt_is_import = stmt_name == Tree.IMPORT_STMT
            self.walk_node(n)

    def walk_if_stmt(self, node: AstType) -> None:
        """AST: if_stmt"""
        self.write_token("if")
        self.write_condition_and_body(node)
        while self.get(node, Token.ELIF):
            self.write_token("elif")
            self.write_condition_and_body(node)
        if self.get(node, Token.ELSE):
            self.write_token("else:")
            self.write_stmt(self.get(node, Tree.EXECUTION_BLOCK))

    def walk_assert_stmt(self, node: AstType) -> None:
        """Syntax
        assert_stmt: ASSERT simple_expr (IF simple_expr)? (COMMA test)?
        """
        self.write_token("assert")
        self.write_token_separator()
        self.write_expr(self.get(node, Tree.SIMPLE_EXPR))
        if self.get(node, Token.IF):
            self.write_token_separator()
            self.write_token("if")
            self.write_token_separator()
            self.write_expr(self.get(node, Tree.SIMPLE_EXPR))
        test_node = self.get(node, Tree.TEST)
        if test_node:
            self.write_token(", ")
            self.write_expr(test_node)

    def walk_schema_stmt(self, node: AstType) -> None:
        """AST: schema_stmt

        schema_stmt: [decorators] (SCHEMA|MIXIN|PROTOCOL) [RELAXED] NAME [LEFT_BRACKETS [schema_arguments] RIGHT_BRACKETS] [LEFT_PARENTHESES identifier (COMMA identifier)* RIGHT_PARENTHESES] COLON NEWLINE [schema_body]
        schema_body: _INDENT (string NEWLINE)* [mixin_stmt] (schema_attribute_stmt|schema_init_stmt)* [check_block] _DEDENT
        schema_attribute_stmt: attribute_stmt NEWLINE
        attribute_stmt: [decorators] (FINAL)? NAME COLON type [(ASSIGN|COMP_OR) test]
        schema_init_stmt: if_simple_stmt | if_stmt
        """

        self.add_blank_line()
        self._is_colon_without_space = True
        keywords = [Token.SCHEMA, Token.MIXIN, Token.PROTOCOL]
        for keyword in keywords:
            if self.has(node, keyword):
                self.walk_nodes(*self.get_internal(node, keyword, Token.COLON))
                break
        self._is_colon_without_space = False
        self.write_expr(node)
        self.add_blank_line()

    def walk_schema_argument(self, node: AstType):
        """AST: schema_argument

        schema_argument: NAME [COLON type] [ASSIGN test]
        """
        nodes = node["children"]
        if self.has(node, Token.COLON):
            self.walk_nodes(*nodes)
        else:
            self.write_arguments(nodes)

    def walk_if_item(self, node: AstType):
        """Syntax
        if_item: IF test COLON if_item_exec_block (ELIF test COLON if_item_exec_block)* (ELSE COLON if_item_exec_block)?
        """
        self._is_in_collection_if = True
        self.write_colon_without_space(node)
        self._is_in_collection_if = False

    def walk_if_entry(self, node: AstType):
        """Syntax
        if_entry: IF test COLON if_entry_exec_block (ELIF test COLON if_entry_exec_block)* (ELSE COLON if_entry_exec_block)?
        """
        self._is_in_collection_if = True
        for n in node["children"]:
            if n["name"] == Token.COLON:
                self.write(":")
            else:
                self.walk_node(n)
        self._is_in_collection_if = False

    # User interfaces

    def fmt_ast(self, ast: AstType) -> str:
        """Format main function and return the format string buffer"""
        if not isinstance(ast, dict):
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CompileError_TYPE, arg_msg="Invalid ast type"
            )
        self.walk_node(ast)
        # Append file last blank line
        if self.last_token_value == ENDLINE_TOKEN:
            self.write(ENDLINE_TOKEN)
        self.printer.seek(0)
        return self.printer.read()[:-1]


def _get_kcl_files(file_or_dir: pathlib.Path, recursively=False):
    """
    Get files in dir
    """
    kcl_file_selector = "*.k"
    return (
        file_or_dir.rglob(kcl_file_selector)
        if recursively
        else file_or_dir.glob(kcl_file_selector)
    )


def kcl_ast_to_fmt_file(json_ast: AstType) -> str:
    """
    Update a kcl ast to a fmt file
    """
    return Formatter().fmt_ast(json_ast)


def kcl_fmt_source(source: str) -> (str, bool):
    """
    Format kcl code string and return formatted string
    """
    formatted_code = Formatter().fmt_ast(get_lark_tree_from_expr(source, False))
    return formatted_code, source != formatted_code


def kcl_fmt_dir(file_or_dir: pathlib.Path, _recursively=False) -> int:
    """
    Format all kcl files in the input directory.
    If 'recursive' is 'True', recursively search kcl files from all its
    sub directories.

    Return
    ------
    Number of formatted files
    """
    return len([kcl_fmt_file(file) for file in _get_kcl_files(file_or_dir)])


def kcl_fmt_file(filepath: pathlib.Path, is_stdout=False) -> bool:
    """
    Format single kcl file
    """
    source, is_formatted = kcl_fmt_source(filepath.read_text(encoding="utf-8"))
    if is_stdout:
        print(source, end="")
    else:
        filepath.write_text(source)
    return is_formatted


def kcl_fmt(input_file_or_path: str, is_stdout=False, recursively=False) -> List[str]:
    """
    Format kcl file or path contains kcl files

    Parameters
    ----------
    - input_file_or_path: Input filename or pathname string
    """
    try:
        changed_paths: List[str] = []
        formatting_filename = None
        path = pathlib.Path(input_file_or_path).resolve()
        if path.is_dir():
            for i, file in enumerate(_get_kcl_files(path, recursively)):
                formatting_filename = file
                if kcl_fmt_file(file, is_stdout):
                    changed_paths.append(str(file.name))
        elif path.is_file():
            formatting_filename = path
            if kcl_fmt_file(path, is_stdout):
                changed_paths.append(str(path.name))
        if not is_stdout:
            format_count = len(changed_paths)
            klog.write_out(
                "KCL format done and {} formatted:\n".format(
                    str(format_count) + " file was"
                    if format_count <= 1
                    else str(format_count) + " files were"
                )
            )
            [klog.write_out(f"{p}\n") for p in changed_paths]
        return changed_paths
    except kcl_error.KCLSyntaxException as err:
        # TODO: Support _Formatter filename context and remove this except.
        # Add filename, line, column and raise
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.InvalidSyntax_TYPE,
            file_msgs=[
                kcl_error.ErrFileMsg(
                    filename=formatting_filename, line_no=err.lineno, col_no=err.colno
                )
            ],
        )

    except Exception as err:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.CompileError_TYPE,
            file_msgs=[
                kcl_error.ErrFileMsg(
                    filename=formatting_filename,
                )
            ],
            arg_msg=str(err),
        )



# Copyright 2021 The KCL Authors. All rights reserved.

import re
from ast import literal_eval
from typing import Union

import kclvm.kcl.ast as ast
import kclvm.compiler.parser as parser
import kclvm.api.object.internal as internal


def BuildNodeFromString(value: str, line: int = None, column: int = None) -> ast.Expr:
    lit_node = BuildLitNodeFromString(value, line, column)
    if isinstance(lit_node, ast.StringLit):
        try:
            val = parser.ParseExpr(code=value)
            # If `val` is a identifier, convert it to a string literal
            return lit_node if isinstance(val, ast.Identifier) else val
        except Exception:
            return lit_node
    return lit_node


def BuildLitNodeFromValue(
    value: Union[int, float, str, bool], line: int = None, column: int = None
) -> ast.Literal:
    if value is None:
        val = ast.NameConstantLit()
        val.value = None
    elif value is internal.Undefined:
        val = ast.NameConstantLit()
        val.value = internal.Undefined
    elif value is True:
        val = ast.NameConstantLit()
        val.value = True
    elif value is False:
        val = ast.NameConstantLit()
        val.value = False
    elif isinstance(value, (int, float)):
        val = ast.NumberLit(value=value)
    else:
        val = ast.StringLit()
        val.value = value if isinstance(value, str) else str(value)
    val.line = line
    val.column = column
    return val


def BuildLitNodeFromString(
    value: str, line: int = None, column: int = None
) -> ast.Literal:
    if value in ["True", "true"]:
        val = ast.NameConstantLit()
        val.value = True
    elif value in ["False", "false"]:
        val = ast.NameConstantLit()
        val.value = False
    elif value in ["None", "null"]:
        val = ast.NameConstantLit()
        val.value = None
    elif value in ["Undefined"]:
        val = ast.NameConstantLit()
        val.value = internal.Undefined
    elif is_number(value):
        val = ast.NumberLit(value=literal_eval(value))
    else:
        val = ast.StringLit()
        val.value = str(value)
        val.raw_value = value

        if val.value and val.value[0] == "'" and val.value[-1] == "'":
            val.value = val.value[1:-1]
        elif val.value and val.value[0] == '"' and val.value[-1] == '"':
            val.value = val.value[1:-1]

        if (
            val.raw_value
            and val.raw_value[0] not in ["'", '"']
            and val.raw_value[-1] not in ["'", '"']
        ):
            val.raw_value = '"' + val.raw_value.replace('"', '\\"') + '"'
    val.line = line
    val.column = column
    return val


def is_number(value: str):
    """Whether a string is a number string"""
    pattern = re.compile(r"^[-+]?[-0-9]\d*\.\d*|[-+]?\.?[0-9]\d*$")
    return bool(pattern.match(value))



# Copyright 2021 The KCL Authors. All rights reserved.

import time
from typing import Optional
from copy import deepcopy

import kclvm.api.object as obj
import kclvm.kcl.error as kcl
import kclvm.kcl.ast as ast
import kclvm.kcl.types as types
import kclvm.config
import kclvm.compiler.extension.builtin

from kclvm.vm.runtime.evaluator.eval import Evaluator
from kclvm.vm.code import Opcode
from kclvm.compiler.check.check_type import (
    type_pack_and_check,
    runtime_types,
)
from kclvm.unification import value_subsume

_evaluator_inst = Evaluator()


def is_kcl_debug():
    return kclvm.config.debug and kclvm.config.verbose > 2


def debug_stack(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    idx = int(arg & 0xFF)
    at = int((arg >> 8) & 0xFF)
    return vm.debug_stack(idx, at)


def debug_locals(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    idx = int(arg & 0xFF)
    at = int((arg >> 8) & 0xFF)
    return vm.debug_locals(idx, at)


def debug_globals(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    idx = int(arg & 0xFF)
    at = int((arg >> 8) & 0xFF)
    return vm.debug_globals(idx, at)


def debug_names(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    idx = int(arg & 0xFF)
    at = int((arg >> 8) & 0xFF)
    return vm.debug_names(idx, at)


def push_frame_locals(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    vm.push_frame_locals()


def pop_frame_locals(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    vm.pop_frame_locals()


def unpack_operand(arg: int):
    return int(arg & 0xFF), int((arg >> 8) & 0xFF), int((arg >> 16) & 0xFF)


def bin_action(vm, code: int, _arg: int) -> Optional[obj.KCLObject]:
    right, left = vm.pop(), vm.top()
    result = _evaluator_inst.eval_binary_op(left, right, code, vm=vm)
    vm.set_top(result)
    return result


def unary_action(vm, code: int, _arg: int) -> Optional[obj.KCLObject]:
    expr_obj = vm.top()
    result = _evaluator_inst.eval_unary_op(expr_obj, code)
    vm.set_top(result)
    return result


def inplace_action(vm, code: int, _arg: int) -> Optional[obj.KCLObject]:
    value, target = vm.pop(), vm.top()
    result = _evaluator_inst.eval_inplace_op(target, value, code, vm=vm)
    vm.set_top(result)
    return result


def bin_add(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_substract(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_floor_divide(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_true_divide(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def build_map(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    config_obj = obj.KCLSchemaConfigObject(value={})
    vm.push(config_obj)
    return config_obj


def load_free(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    name = vm.names[arg]

    # If the variable in local:
    if vm.ctx.locals_list:
        index = len(vm.ctx.locals_list) - 1
        while index >= 0:
            if name in vm.ctx.locals_list[index]:
                vm.push(vm.ctx.locals_list[index][name])
                return vm.ctx.locals_list[index][name]
            index -= 1

    if name in vm.ctx.locals:
        vm.push(vm.ctx.locals[name])
        return vm.ctx.locals[name]

    free_obj = obj.NONE_INSTANCE
    index = vm.frame_index
    # Search the local variables from the inside to the outside schema
    while index >= 1:
        index -= 1
        if name in vm.frames[index].locals:
            free_obj = vm.frames[index].locals[name]

    vm.push(free_obj)
    return free_obj


def load_closure(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    # free_obj = vm.ctx.free_vars[arg]
    closure_obj = (
        vm.ctx.free_vars[arg] if arg < len(vm.ctx.free_vars) else obj.UNDEFINED_INSTANCE
    )
    vm.push(closure_obj)
    return closure_obj


def make_function(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return make_compiled_function(vm, code, arg)


def make_closure(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return make_compiled_function(vm, code, arg, is_closure=True)


def make_compiled_function(
    vm, _code: int, arg: int, is_closure: bool = False
) -> Optional[obj.KCLObject]:
    """
    Pushes a new function object on the stack. From bottom to top, the consumed stack must consist of values if the argument carries a specified flag value
    + 0x01 a tuple of default values for positional-only and positional-or-keyword parameters in positional order
    + 0x02 a dictionary of keyword-only parameters’ default values
    + 0x04 an annotation dictionary
    + 0x08 a tuple containing cells for free variables, making a closure
    + the code associated with the function (at TOS1)
    + the qualified name of the function (at TOS)
    """
    # Pop the function name
    name = vm.pop().value
    # Pop code placeholder
    runtime_code = vm.pop()
    closure = vm.pop() if is_closure else None
    args, _, _ = unpack_operand(arg)
    # var params []grumpy.Param
    params: list = [0] * args
    index = 0
    for index in range(args):
        # positional args placeholders to fit grumpy arg validate
        arg_value = vm.pop()
        arg_type = vm.pop().value
        arg_name = vm.pop().value
        params[index] = obj.Parameter(
            name=arg_name, type_annotation=arg_type, value=arg_value
        )

    func = obj.KCLCompiledFunctionObject(
        name=name,
        instructions=runtime_code.codes,
        names=runtime_code.names,
        constants=runtime_code.constants,
        num_locals=0,
        num_parameters=arg,
        params=params[::-1],
        pkgpath=vm.cur_run_pkg,
        closure=closure,
    )
    vm.push(func)
    return func


def pop_top(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    return vm.pop()


def rot_two(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    v = vm.pop()
    w = vm.pop()
    vm.push(v)
    vm.push(w)
    return w


def rot_three(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    v = vm.pop()
    w = vm.pop()
    u = vm.pop()
    vm.push(v)
    vm.push(w)
    vm.push(u)
    return u


def dup_top(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    vm.push(vm.peek())
    return vm.peek()


def copy_top(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    vm.push(deepcopy(vm.peek()))
    return vm.peek()


def dup_top_two(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    v = vm.pop()
    w = vm.pop()
    vm.push(w)
    vm.push(v)
    vm.push(w)
    vm.push(v)
    return v


def nop(_vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    """Empty opcode action"""
    pass


def unary_pos(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return unary_action(vm, code, arg)


def unary_neg(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return unary_action(vm, code, arg)


def unary_not(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return unary_action(vm, code, arg)


def unary_invert(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return unary_action(vm, code, arg)


def bin_power(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_mul(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_mod(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_subscr(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def inplace_floor_div(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def inplace_true_div(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def store_map(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    v, k = vm.pop(), vm.pop()
    config_ref = vm.peek()
    if v.type() == obj.KCLObjectType.UNPACK and k.type() in [
        obj.KCLObjectType.NONE,
        obj.KCLObjectType.UNDEFINED,
    ]:
        config_ref.append_unpack(v.unpack())
    else:
        config_ref.update_key_value(k, v)
    return config_ref


def inplace_add(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def inplace_sub(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def inplace_mul(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def inplace_mod(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def store_subscr(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    """v[w] = u"""
    w = vm.pop()
    v = vm.pop()
    u = vm.pop()
    _evaluator_inst.set_item(v, w, u)
    return v


def bin_lshift(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_rshift(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_and(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_xor(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_or(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_logic_and(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def bin_logic_or(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return bin_action(vm, code, arg)


def inplace_pow(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def get_iter(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    assert 0 < _arg <= 2
    iter_obj = obj.KCLIterObject.build_iter(vm.top(), _arg)
    vm.set_top(iter_obj)
    return None


def print_expr(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    expr_obj = vm.pop()
    print(expr_obj.value)
    return expr_obj


def emit_expr(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    emitted_obj = vm.pop()
    if isinstance(emitted_obj, obj.KCLSchemaObject):
        output_type = (
            emitted_obj.get(obj.SCHEMA_SETTINGS_ATTR_NAME)
            .get(obj.SETTINGS_OUTPUT_KEY)
            .value
        )
        if (
            output_type == obj.SETTINGS_OUTPUT_STANDALONE
            or output_type == obj.SETTINGS_OUTPUT_IGNORE
        ):
            # Internal magic variable
            vm.update_global(f"$EMIT_VAR_{time.time()}", emitted_obj)
    return emitted_obj


def inplace_lshift(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def inplace_rlshift(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def inplace_and(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def inplace_xor(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def inplace_or(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    return inplace_action(vm, code, arg)


def return_value(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    vm.pop_frame()
    return vm.peek()


def return_last_value(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    frame = vm.pop_frame()
    variables = list(frame.locals.values())
    return_obj = variables[-1] if variables else obj.NONE_INSTANCE
    vm.push(return_obj)
    return return_obj


def store_name(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    return vm.store_name(arg)


def unpack_sequence(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    sequence_obj = vm.pop()
    unpack_obj = obj.KCLUnpackObject(sequence_obj, arg != 1)
    vm.push(unpack_obj)
    return unpack_obj


def for_iter(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    """
    TOS is an iterator. Call its iter_next() method. If this yields a new value,
    push it on the stack (leaving the iterator below it).
    If the iterator indicates it is exhausted TOS is popped,
    and the byte code counter is incremented by 'arg' argument.
    """
    try:
        iter_next_obj = _evaluator_inst.iter_next(vm.peek())
        assert isinstance(iter_next_obj, obj.KCLTupleObject)
        # Push loop variables
        for o in iter_next_obj.value[::-1]:
            vm.push(obj.to_kcl_obj(o))
    except StopIteration:
        it = vm.pop()
        assert isinstance(it, obj.KCLIterObject)
        vm.set_instruction_pointer(arg)
    except Exception as err:
        kcl.report_exception(
            err_type=kcl.ErrType.EvaluationError_TYPE, arg_msg=str(err)
        )

    return None


def store_attr(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    """v.w = u"""
    w = vm.names[arg]
    v = vm.pop()
    u = vm.pop()

    _evaluator_inst.set_attr(v, obj.KCLStringObject(w), u, vm=vm)

    return v


def store_global(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    return vm.store_name(arg)


def load_const(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    return vm.load_const(arg)


def load_name(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    return vm.load_name(arg)


def build_list(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    items = [vm.pop() for _i in range(arg)][::-1]
    list_obj = obj.KCLListObject(items=[])
    for item in items:
        if item.type() == obj.KCLObjectType.UNPACK:
            list_obj.append_unpack(item.unpack())
        else:
            list_obj.append(item)
    vm.push(list_obj)
    return list_obj


def schema_attr(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    """
    Define a schema attr into schema type object
    """
    n_decorators = arg
    decorators_obj = [vm.pop() for _i in range(n_decorators)][::-1]
    expected_types = [vm.pop().value]
    attr, value, has_default, is_final, is_optional, op_code = (
        vm.pop().value,
        vm.pop(),
        vm.pop().value,
        vm.pop().value,
        vm.pop().value,
        vm.pop().value,
    )
    config_meta = vm.lazy_eval_ctx.config_meta
    config_obj = vm.ctx.locals[obj.SCHEMA_CONFIG_VALUE_KEY]
    schema_obj = vm.ctx.locals[obj.SCHEMA_SELF_VALUE_KEY]
    config_value = config_obj.get(attr)
    schema_obj.set_attr_type(attr, expected_types)
    schema_obj.set_attr_optional(attr, is_optional)
    attr_runtime_types = runtime_types(expected_types, vm=vm)
    if attr_runtime_types:
        schema_obj.set_attr_runtime_type(attr, attr_runtime_types)
    for decorator in decorators_obj:
        schema_obj.add_decorator(attr, decorator)

    operation = (
        config_obj.get_operation(attr)
        if isinstance(config_obj, obj.KCLSchemaConfigObject)
        else ast.ConfigEntryOperation.UNION
    )
    insert_index = (
        config_obj.get_insert_index(attr)
        if isinstance(config_obj, obj.KCLSchemaConfigObject)
        else None
    )

    # Update and union with config_value
    if has_default:
        target = (
            schema_obj.get(attr) if attr in schema_obj else obj.KCLNoneObject.instance()
        )
        if op_code is None:
            op_value = value
        else:
            op_value = _evaluator_inst.eval_inplace_op(target, value, op_code, vm=vm)
        op_value = type_pack_and_check(op_value, expected_types, vm=vm)
        if schema_obj.get_immutable_flag(attr) and target.value != op_value.value:
            kcl.report_exception(
                err_type=kcl.ErrType.ImmutableRuntimeError_TYPE,
                arg_msg=f"final schema field '{attr}'",
            )
        vm.lazy_eval_ctx.set_value(attr, op_value)
    elif attr not in schema_obj:
        schema_obj.update({attr: obj.KCLUndefinedObject.instance()})

    if attr in config_obj:
        conf_meta = config_meta.get(attr, {})
        filename = conf_meta.get("filename")
        lineno = conf_meta.get("lineno")
        columnno = conf_meta.get("columnno")
        if is_final and obj.to_python_obj(config_value) != obj.to_python_obj(
            schema_obj.attrs.get(attr)
        ):
            kcl.report_exception(
                err_type=kcl.ErrType.ImmutableRuntimeError_TYPE,
                file_msgs=[
                    kcl.ErrFileMsg(filename=filename, line_no=lineno, col_no=columnno)
                ],
                arg_msg=f"final schema field '{attr}'",
            )
        _evaluator_inst.update_schema_attr(
            attr=attr,
            schema_obj=schema_obj,
            config_value=config_value,
            conf_meta=conf_meta,
            expected_types=expected_types,
            operation=operation,
            index=insert_index,
            vm=vm,
            filename=filename,
            lineno=lineno,
            columnno=columnno,
        )
        vm.lazy_eval_ctx.set_value(attr, schema_obj.get(attr))

    # Update mutable flag
    schema_obj.set_immutable_flag(attr, is_final)
    vm.update_local(attr, schema_obj.get(attr))
    return schema_obj


def schema_update_attr(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    name = vm.pop().value
    value = vm.pop()

    schema_obj = vm.ctx.locals[obj.SCHEMA_SELF_VALUE_KEY]
    config_obj = vm.ctx.locals[obj.SCHEMA_CONFIG_VALUE_KEY]
    config_meta = vm.lazy_eval_ctx.config_meta
    config_value = config_obj.get(name)

    # A private schema attribute can be added into schema object whether it's relaxed or not
    if schema_obj.get_immutable_flag(name):
        kcl.report_exception(
            err_type=kcl.ErrType.ImmutableRuntimeError_TYPE,
            arg_msg=f"final schema field '{name}'",
        )

    conf_meta = config_meta.get(name, {})
    filename = conf_meta.get("filename")
    lineno = conf_meta.get("lineno")
    columnno = conf_meta.get("columnno")
    if name not in schema_obj or not schema_obj.get_attr_type(name):
        schema_obj.set_attr_optional(name, True)
    vm.lazy_eval_ctx.set_value(
        name, type_pack_and_check(value, schema_obj.get_attr_type(name), vm=vm)
    )

    if name in config_obj:
        config_value = type_pack_and_check(
            config_value,
            schema_obj.get_attr_type(name),
            vm=vm,
            filename=filename,
            lineno=lineno,
            columno=columnno,
            config_meta=conf_meta.get("$conf_meta"),
        )
        cfg_obj = obj.KCLSchemaConfigObject(value={name: config_value})
        cfg_obj.update_attr_op_using_obj(config_obj)
        schema_obj.union_with(cfg_obj)
        vm.lazy_eval_ctx.set_value(name, schema_obj.get(name))


def make_decorator(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    decorator = vm.pop()
    assert isinstance(decorator, obj.KCLDecoratorObject)
    args, kwargs = _evaluator_inst.call_vars_and_keywords(arg, vm)
    decorator.resolve(args, kwargs)
    vm.push(decorator)
    return decorator


def schema_load_attr(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    name_obj = vm.top()
    vm.set_top(vm.lazy_eval_ctx.get_value(name_obj.value))
    return name_obj


def import_name(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    return vm.import_name(_code, arg)


def jump_forward(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    vm.set_instruction_pointer(arg)
    return None


def jump_if_false_or_pop(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    condition_obj = vm.peek()
    if not obj.KCLObject.is_truthy(condition_obj):
        vm.set_instruction_pointer(arg)
    else:
        vm.pop()
    return None


def jump_if_true_or_pop(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    if obj.KCLObject.is_truthy(vm.peek()):
        vm.set_instruction_pointer(arg)
    else:
        vm.pop()
    return None


def jump_absolute(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    vm.set_instruction_pointer(arg)
    return None


def pop_jump_if_false(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    condition_obj = vm.pop()
    if not obj.KCLObject.is_truthy(condition_obj):
        vm.set_instruction_pointer(arg)
    return None


def pop_jump_if_true(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    condition_obj = vm.pop()
    if obj.KCLObject.is_truthy(condition_obj):
        vm.set_instruction_pointer(arg)
    return None


def load_global(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    vm.load_name(arg)
    return None


def raise_varargs(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    if arg != 1:
        raise Exception(f"invalid raise_varargs opcode arg {arg}")
    msg_obj = vm.pop()
    msg = msg_obj.value if msg_obj.type() == obj.KCLObjectType.STRING else ""
    if arg == 1:
        kcl.report_exception(
            err_type=kcl.ErrType.AssertionError_TYPE, arg_msg=msg if msg else ""
        )
    return None


def raise_check(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    if arg != 1:
        raise Exception(f"invalid raise_check opcode arg {arg}")
    msg_obj = vm.pop()
    msg = msg_obj.value if msg_obj.type() == obj.KCLObjectType.STRING else ""
    if arg == 1:
        config_meta = obj.to_python_obj(vm.ctx.locals[obj.SCHEMA_CONFIG_META_KEY])
        conf_filename, conf_line, conf_column = (
            config_meta.get("$filename"),
            config_meta.get("$lineno"),
            config_meta.get("$columnno"),
        )
        filename, line, _ = vm.get_info(True)
        kcl.report_exception(
            err_type=kcl.ErrType.SchemaCheckFailure_TYPE,
            file_msgs=[
                kcl.ErrFileMsg(
                    filename=filename,
                    line_no=line,
                    arg_msg=kcl.SCHEMA_CHECK_FILE_MSG_COND,
                ),
                kcl.ErrFileMsg(
                    filename=conf_filename,
                    line_no=conf_line,
                    col_no=conf_column,
                    arg_msg=kcl.SCHEMA_CHECK_FILE_MSG_ERR,
                ),
            ],
            arg_msg=msg if msg else "",
        )
    return None


def load_local(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    vm.load_local(arg)
    return None


def store_local(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    vm.store_local(arg)
    return None


def call_function(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    _evaluator_inst.eval_call(code, arg, vm)
    return None


def build_slice(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    if arg != 2 and arg != 3:
        raise Exception(f"invalid slice argc {arg}")
    step = obj.KCLNoneObject.instance()
    if arg == 3:
        step = vm.pop()
    stop = vm.pop()
    start = vm.top()
    vm.set_top(obj.KCLSliceObject(start=start, stop=stop, step=step))
    return None


def list_append(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    item = vm.pop()
    list_obj = vm.peek_nth(arg)
    _evaluator_inst.list_append(list_obj, item)
    return list_obj


def map_add(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    operation = vm.pop().value
    key = vm.pop()
    value = vm.pop()
    config_ref = vm.peek_nth(arg)
    data_obj = obj.KCLSchemaConfigObject(value={key.value: value})
    if operation is None or operation == ast.ConfigEntryOperation.UNION:
        config_ref.union_with(data_obj)
    elif operation == ast.ConfigEntryOperation.OVERRIDE:
        config_ref.update(data_obj)
    elif operation == ast.ConfigEntryOperation.INSERT:
        config_ref.insert_with(data_obj, -1)
    else:
        config_ref.update(data_obj)
    config_ref.add_operation(key.value, operation)
    return config_ref


def delete_item(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    collection_obj = vm.peek_nth(arg)
    is_two_var = vm.pop().value
    value = vm.pop()
    item = vm.pop()
    if isinstance(collection_obj, (obj.KCLSchemaObject, obj.KCLDictObject)):
        collection_obj.delete(item)
    elif isinstance(collection_obj, obj.KCLListObject):
        if is_two_var:
            collection_obj.remove(value)
        else:
            collection_obj.remove(item)
    else:
        kcl.report_exception(
            err_type=kcl.ErrType.EvaluationError_TYPE,
            arg_msg=f"illegal quantifier expression type '{collection_obj.type_str()}'",
        )
    return collection_obj


def schema_nop(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    """Schema nop between scheam update attributes"""
    schema_obj = vm.ctx.locals[obj.SCHEMA_SELF_VALUE_KEY]
    return schema_obj


def make_schema(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    """
    Stack Layout
    ------------
    TOS
        - 6. index signature
        - 5. decorator list
        - 4. check func
        - 3. schema_body_func
        - 2. mixin type object list
        - 1. parent_type_obj
        - 0. schema_type_obj
    """
    decorator_count, mixin_count, _ = unpack_operand(arg)
    index_signature = None
    index_signature_key_type = vm.pop().value
    if index_signature_key_type:
        index_signature = obj.KCLSchemaIndexSignatureObject(
            key_type=index_signature_key_type,
            value_type=vm.pop().value,
            key_name=vm.pop().value,
            any_other=vm.pop().value,
            value=vm.pop(),
        )
    decorators = [vm.pop() for _i in range(decorator_count)][::-1]
    schema_check_func = vm.pop()
    schema_body_func = vm.pop()
    mixins = [vm.pop() for _i in range(mixin_count)][::-1]
    parent_schema_type = vm.pop()
    if parent_schema_type.type() not in [
        obj.KCLObjectType.SCHEMA_TYPE,
        obj.KCLObjectType.NONE,
        obj.KCLObjectType.UNDEFINED,
    ]:
        kcl.report_exception(
            err_type=kcl.ErrType.EvaluationError_TYPE,
            arg_msg="illegal schema inherit object type",
        )
    self_schema_type = vm.pop()

    type_obj = obj.KCLSchemaTypeObject.new(
        self_schema_type.name,
        parent_schema_type
        if isinstance(parent_schema_type, obj.KCLSchemaTypeObject)
        else None,
        self_schema_type.protocol,
        filename=vm.get_info()[0],
        is_mixin=self_schema_type.is_mixin,
        pkgpath=self_schema_type.pkgpath,
        attr_list=self_schema_type.attr_list,
        index_signature=index_signature,
        is_relaxed=bool(index_signature),
        vm=vm,
    )

    vm.define_schema_type(
        f"{self_schema_type.pkgpath}.{self_schema_type.name}", type_obj
    )

    if isinstance(schema_body_func, obj.KCLCompiledFunctionObject):
        schema_body_func.name = self_schema_type.name
        type_obj.set_func(schema_body_func)
    else:
        assert isinstance(schema_body_func, (obj.KCLNoneObject, obj.KCLUndefinedObject))

    if isinstance(schema_check_func, obj.KCLCompiledFunctionObject):
        schema_check_func.name = self_schema_type.name
        type_obj.set_check_func(schema_check_func)
    else:
        assert isinstance(
            schema_check_func, (obj.KCLNoneObject, obj.KCLUndefinedObject)
        )

    type_obj.add_decorators(decorators)

    if len(mixins) > 0:
        if isinstance(mixins[0], obj.KCLStringObject):
            type_obj.mixins_names = [x.value for x in mixins]
        elif isinstance(mixins[0], obj.KCLSchemaTypeObject):
            type_obj.mixins = mixins
        else:
            assert False, f"make_schema: mixins[0] type: {type(mixins[0])}"
    type_obj.update_mixins(vm=vm)
    vm.push(type_obj)
    return type_obj


def build_schema(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    schema_obj = vm.pop()
    config_obj = vm.pop()
    config_meta = obj.to_python_obj(vm.pop())
    finalize = vm.pop().value
    args, kwargs = _evaluator_inst.call_vars_and_keywords(arg, vm)
    # Schema type object
    if isinstance(schema_obj, obj.KCLSchemaTypeObject):
        if isinstance(config_obj, obj.KCLDictObject):
            inst = schema_obj.new_instance(
                config_obj, config_meta, args, kwargs, vm, finalize=finalize
            )
        else:
            inst = config_obj
    # Schema value object
    else:
        inst = _evaluator_inst.eval_binary_op(
            schema_obj, config_obj, Opcode.BINARY_OR, vm=vm
        )
    vm.push(inst)
    return inst


def build_schema_config(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    config_obj = obj.KCLSchemaConfigObject(
        value={}, operation_map={}, insert_index_map={}
    )
    vm.push(config_obj)
    return config_obj


def store_schema_config(vm, _code: int, _arg: int) -> Optional[obj.KCLObject]:
    insert_index, operation, is_nest_key = (
        vm.pop().value,
        vm.pop().value,
        vm.pop().value,
    )
    v, k = vm.pop(), vm.pop()
    config_ref = vm.peek()
    if v.type() == obj.KCLObjectType.UNPACK and k.type() in [
        obj.KCLObjectType.NONE,
        obj.KCLObjectType.UNDEFINED,
    ]:
        config_ref.append_unpack(v.unpack())
    else:
        # Deal the nest var config e.g., {a.b.c: "123"} -> {a: {b: {c: "123"}}}
        if k.type() == obj.KCLObjectType.STRING and is_nest_key:
            data_obj = obj.KCLSchemaConfigObject()
            obj_ref = data_obj
            nest_keys = k.value.split(".")
            for i, key in enumerate(nest_keys):
                obj_ref.value[key] = (
                    v if i == len(nest_keys) - 1 else obj.KCLSchemaConfigObject()
                )
                if i == len(nest_keys) - 1:
                    obj_ref.add_operation(key, operation, insert_index)
                obj_ref = obj_ref.value[key]
            operation = ast.ConfigEntryOperation.UNION
            insert_index = None
        else:
            data_obj = obj.KCLSchemaConfigObject(value={k.value: v})
            config_ref.add_operation(k.value, operation, insert_index)
        if operation is None or operation == ast.ConfigEntryOperation.UNION:
            config_ref.union_with(data_obj)
        elif operation == ast.ConfigEntryOperation.OVERRIDE:
            config_ref.update(data_obj)
        elif operation == ast.ConfigEntryOperation.INSERT:
            config_ref.insert_with(data_obj, insert_index)
        elif operation == ast.ConfigEntryOperation.UNIQUE:
            config_ref.unique_merge_with(data_obj)
        elif operation == ast.ConfigEntryOperation.UNIFICATION:
            if value_subsume(data_obj, config_ref):
                config_ref.union_with(data_obj)
            else:
                kcl.report_exception(
                    err_type=kcl.ErrType.EvaluationError_TYPE,
                    arg_msg="unification conflict",
                )
        else:
            config_ref.union_with(data_obj)
    return config_ref


def load_attr(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    attr_obj = vm.top()
    name = vm.names[arg]
    value = _evaluator_inst.load_attr(attr_obj, name)
    vm.set_top(value)
    return value


def load_builtin(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    return vm.load_builtin(arg)


def format_values(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    assert arg >= 0
    formatted_str_obj_list = []
    format_spec = vm.pop().value
    for i in range(arg):
        format_obj = vm.pop()
        format_value = _evaluator_inst.format_value(format_obj, format_spec)
        formatted_str_obj_list.append(format_value)
    for str_obj in formatted_str_obj_list[::-1]:
        vm.push(str_obj)
    return vm.peek()


def member_ship_as(vm, code: int, arg: int) -> Optional[obj.KCLObject]:
    # Pop the type object
    type_object = vm.pop()
    vm.set_top(types.type_convert(vm.top(), type_object))
    return vm.top()


def compare_op(vm, _code: int, arg: int) -> Optional[obj.KCLObject]:
    left, right = vm.pop(), vm.top()
    result = _evaluator_inst.eval_compare_op(left, right, arg, vm=vm)
    vm.set_top(result)
    return result


VM_OP_ACTIONS = {
    Opcode.BINARY_ADD: bin_add,
    Opcode.BINARY_SUBTRACT: bin_substract,
    Opcode.BINARY_MULTIPLY: bin_mul,
    Opcode.BINARY_FLOOR_DIVIDE: bin_floor_divide,
    Opcode.BINARY_TRUE_DIVIDE: bin_true_divide,
    Opcode.BUILD_SCHEMA_CONFIG: build_schema_config,
    Opcode.STORE_SCHEMA_CONFIG: store_schema_config,
    Opcode.POP_TOP: pop_top,
    Opcode.ROT_TWO: rot_two,
    Opcode.ROT_THREE: rot_three,
    Opcode.DUP_TOP: dup_top,
    Opcode.DUP_TOP_TWO: dup_top_two,
    Opcode.COPY_TOP: copy_top,
    Opcode.NOP: nop,
    Opcode.UNARY_POSITIVE: unary_pos,
    Opcode.UNARY_NEGATIVE: unary_neg,
    Opcode.UNARY_NOT: unary_not,
    Opcode.UNARY_INVERT: unary_invert,
    Opcode.BINARY_POWER: bin_power,
    Opcode.BINARY_MODULO: bin_mod,
    Opcode.BINARY_SUBSCR: bin_subscr,
    Opcode.INPLACE_FLOOR_DIVIDE: inplace_floor_div,
    Opcode.INPLACE_TRUE_DIVIDE: inplace_true_div,
    Opcode.STORE_MAP: store_map,
    Opcode.INPLACE_ADD: inplace_add,
    Opcode.INPLACE_SUBTRACT: inplace_sub,
    Opcode.INPLACE_MULTIPLY: inplace_mul,
    Opcode.INPLACE_MODULO: inplace_mod,
    Opcode.STORE_SUBSCR: store_subscr,
    Opcode.BINARY_LSHIFT: bin_lshift,
    Opcode.BINARY_RSHIFT: bin_rshift,
    Opcode.BINARY_AND: bin_and,
    Opcode.BINARY_XOR: bin_xor,
    Opcode.BINARY_OR: bin_or,
    Opcode.BINARY_LOGIC_AND: bin_logic_and,
    Opcode.BINARY_LOGIC_OR: bin_logic_or,
    Opcode.INPLACE_POWER: inplace_pow,
    Opcode.GET_ITER: get_iter,
    Opcode.PRINT_EXPR: print_expr,
    Opcode.EMIT_EXPR: emit_expr,
    Opcode.INPLACE_LSHIFT: inplace_lshift,
    Opcode.INPLACE_RSHIFT: inplace_rlshift,
    Opcode.INPLACE_AND: inplace_and,
    Opcode.INPLACE_XOR: inplace_xor,
    Opcode.INPLACE_OR: inplace_or,
    Opcode.RETURN_VALUE: return_value,
    Opcode.RETURN_LAST_VALUE: return_last_value,
    Opcode.STORE_NAME: store_name,
    Opcode.UNPACK_SEQUENCE: unpack_sequence,
    Opcode.FOR_ITER: for_iter,
    Opcode.STORE_ATTR: store_attr,
    Opcode.STORE_GLOBAL: store_global,
    Opcode.LOAD_CONST: load_const,
    Opcode.LOAD_NAME: load_name,
    Opcode.BUILD_LIST: build_list,
    Opcode.BUILD_MAP: build_map,
    Opcode.LOAD_ATTR: load_attr,
    Opcode.IMPORT_NAME: import_name,
    Opcode.JUMP_FORWARD: jump_forward,
    Opcode.JUMP_IF_FALSE_OR_POP: jump_if_false_or_pop,
    Opcode.JUMP_IF_TRUE_OR_POP: jump_if_true_or_pop,
    Opcode.JUMP_ABSOLUTE: jump_absolute,
    Opcode.POP_JUMP_IF_FALSE: pop_jump_if_false,
    Opcode.POP_JUMP_IF_TRUE: pop_jump_if_true,
    Opcode.LOAD_GLOBAL: load_global,
    Opcode.RAISE_VARARGS: raise_varargs,
    Opcode.RAISE_CHECK: raise_check,
    Opcode.LOAD_LOCAL: load_local,
    Opcode.STORE_LOCAL: store_local,
    Opcode.LOAD_FREE: load_free,
    Opcode.CALL_FUNCTION: call_function,
    Opcode.MAKE_FUNCTION: make_function,
    Opcode.BUILD_SLICE: build_slice,
    Opcode.LOAD_CLOSURE: load_closure,
    Opcode.MAKE_CLOSURE: make_closure,
    Opcode.LIST_APPEND: list_append,
    Opcode.MAP_ADD: map_add,
    Opcode.DELETE_ITEM: delete_item,
    Opcode.MAKE_SCHEMA: make_schema,
    Opcode.BUILD_SCHEMA: build_schema,
    Opcode.SCHEMA_ATTR: schema_attr,
    Opcode.LOAD_BUILT_IN: load_builtin,
    Opcode.COMPARE_OP: compare_op,
    Opcode.MAKE_DECORATOR: make_decorator,
    Opcode.SCHEMA_LOAD_ATTR: schema_load_attr,
    Opcode.SCHEMA_UPDATE_ATTR: schema_update_attr,
    Opcode.SCHEMA_NOP: schema_nop,
    Opcode.FORMAT_VALUES: format_values,
    Opcode.MEMBER_SHIP_AS: member_ship_as,
    Opcode.PUSH_FRAME_LOCALS: push_frame_locals,
    Opcode.POP_FRAME_LOCALS: pop_frame_locals,
    Opcode.DEBUG_STACK: debug_stack,
    Opcode.DEBUG_LOCALS: debug_locals,
    Opcode.DEBUG_GLOBALS: debug_globals,
    Opcode.DEBUG_NAMES: debug_names,
}



# Copyright 2021 The KCL Authors. All rights reserved.

from .util import dotdict, hash, merge_option_same_keys, safe_call
from .check_utils import (
    check_allow_none,
    check_all_allow_none,
    check_not_none,
    check_all_not_none,
    PreCheck,
    PostCheck,
    CheckRules,
    CHECK_MODE,
    alert_internal_bug,
    check_type_not_none,
    check_type_allow_none,
)

__all__ = [
    "dotdict",
    "hash",
    "merge_option_same_keys",
    "check_allow_none",
    "check_all_allow_none",
    "check_not_none",
    "check_all_not_none",
    "PreCheck",
    "PostCheck",
    "CheckRules",
    "CHECK_MODE",
    "alert_internal_bug",
    "check_type_not_none",
    "check_type_allow_none",
    "safe_call",
]



# class DocTemplater
import os
import kclvm.tools.docs.model_pb2 as model


def md_schema_section_render(schema_name: str):
    return f"Schema {schema_name}"


def md_module_doc_templater(module: model.ModuleDoc) -> str:
    def table(records, headings) -> str:
        """
        Generate a Markdown table from records.
        records -- Iterable.  Rows will be generated from this.
        headings -- List of column headings.
        """
        table_heading = (
            f"|{'|'.join(headings)}|\n|{'|'.join(['-' * len(h) for h in headings])}|\n"
        )
        table_rows = "\n".join(
            [f"|{record[0]}|{record[1]}|{record[2]}|{record[3]}|" for record in records]
        )
        return f"{table_heading}{table_rows}\n"

    # module header
    header = f"# {module.name}\n\n"
    # source code link
    file_name = module.relative_path.replace(f".{os.sep}", "")
    source_code = (
        f"Source: [{file_name}]({module.source_code_url}{os.sep}{file_name})\n\n"
        if module.source_code_url
        else ""
    )
    # module doc string
    module_doc = f"{module.doc}\n\n" if module.doc else ""
    # schemas header
    schemas_section = ""
    if module.schemas:
        for schema in module.schemas:
            _schema_header = f"## Schema {schema.name}\n\n"
            _schema_docstring = f"{schema.doc}\n\n" if schema.doc else ""

            _base_schema_section = ""
            if schema.base_schema and schema.base_schema.type_str:
                _base_schema_section = (
                    f"### Base Schema\n{schema.base_schema.type_str}\n\n"
                )

            attributes_section = ""
            if schema.attributes:
                attributes_header = "### Attributes\n\n"
                attributes_table = table(
                    [
                        [
                            f"**{a.name}**<br />{a.doc}" if a.doc else f"**{a.name}**",
                            a.type.type_str,
                            a.default_value if a.default_value else "Undefined",
                            "optional" if a.is_optional else "**required**",
                        ]
                        for a in schema.attributes
                    ],
                    [
                        "Name and Description",
                        "Type",
                        "Default Value",
                        "Required",
                    ],
                )
                attributes_section = f"{attributes_header}{attributes_table}"
            # Examples
            _examples = (
                f"### Examples\n```python\n{schema.examples}\n```\n\n"
                if schema.examples
                else ""
            )
            schemas_section = f"{schemas_section}{_schema_header}{_schema_docstring}{_base_schema_section}{attributes_section}{_examples}"
    # not editable footer
    not_editable = "<!-- Auto generated by kcl-doc tool, please do not edit. -->"
    return f"{header}{source_code}{module_doc}{schemas_section}{not_editable}"



"""
Docstring parser for KCL models.

Basically, we follow the docstring standard of numpy doc, use the related tools as underlying libs
and extend the part which it does not cover, such as default value.

Reference:
+ https://numpydoc.readthedocs.io/
"""

import re
import typing
from numpydoc.docscrape import NumpyDocString

import kclvm.kcl.ast as ast
import kclvm.api.object as objpkg
import kclvm.tools.docs.checker as checker
import kclvm.tools.docs.model_pb2 as model

# Try to match type, optional, and default value.
# Note: there is no standard way to define default value in numpydoc, we define it as below.
TYPE_OPTIONAL_DEFAULT_REGEX = re.compile(
    r"(?P<type>.*?)(,?\s*[Dd]efault(?: is | = |: |s to |)\s*(?P<value>.*?))?(,?\s*(?P<optional>optional|\(optional\)|required|\(required\)).?)?$"
)


class SchemaDocParser:
    """Schema doc parser.

    The schema doc is defined in model.SchemaDoc

    :param schema. The schema ast.
    :param doc. The schema docstring
    """

    def __init__(
        self,
        schema: ast.SchemaStmt,
        schema_type: objpkg.KCLSchemaTypeObject,
        doc=None,
        root: str = None,
        checker=checker.SchemaDocStringChecker,
    ):
        self._schema = schema
        self.root = root
        self.schema_type = schema_type

        if doc is None:
            if schema is None:
                raise ValueError("No schema or documentation string given")

            import inspect

            doc = inspect.cleandoc(schema.doc)
        schema_doc = NumpyDocString(doc)

        # Base Schema relation
        base_schema: model.Type = None
        if self._schema.parent_name:
            assert self.schema_type.base
            base_schema = model.Type(
                type_str=self._schema.parent_name.get_name(),
                type_category=model.Type.TypeCategory.SCHEMA,
                schema_type=model.SchemaType(
                    name=self.schema_type.base.name,
                    relative_path=self.schema_type.base.filename.replace(
                        self.root, ".", 1
                    ),
                ),
            )

        doc = model.SchemaDoc(name=schema.name, base_schema=base_schema)

        # Summary
        summary = ""
        for sum_line in schema_doc["Summary"]:
            summary += sum_line + "\n"
        doc.doc = summary

        # Attributes
        # From docstring: collect the attribute map from the Attributes docstring
        doc_attr_map: typing.Dict[str, model.SchemaAttributeDoc] = {}
        for attr in schema_doc["Attributes"]:
            doc_attr_name = attr[0]
            doc_attr_type = attr[1]
            doc_attr_desc = ""
            doc_attr_default = None
            doc_attr_optional = False

            if attr[2]:
                for desc in attr[2]:
                    doc_attr_desc += desc + "\n"

            # To prevent the doc pattern e.g., `name: type`, and modify it with kcl format tool
            if ":" in doc_attr_name:
                name_type_parts = doc_attr_name.split(":", 1)
                if name_type_parts and len(name_type_parts) >= 2:
                    doc_attr_name = name_type_parts[0].strip()
                    doc_attr_type = name_type_parts[1].strip() + doc_attr_type
            doc_attr_name = doc_attr_name.lstrip("$")
            if doc_attr_type:
                match = TYPE_OPTIONAL_DEFAULT_REGEX.match(doc_attr_type)
                if (
                    match is None
                    or match.group(0) == match.group(2)
                    or match.group(0) == match.group(3)
                ):
                    # if there is no type defined, optional(group2) or default(group3)
                    # may be treated as type(group0)
                    # fixme: hacky, find a better way to check type
                    doc_attr_type = ""
                else:
                    doc_attr_type = match.group("type")
                    # reset on all false cases
                    doc_attr_type = "" if not doc_attr_type else doc_attr_type
                if match is not None:
                    doc_attr_default = match.group("value")
                    optional = match.group("optional")
                    if optional is not None:
                        doc_attr_optional = "optional" in optional
            doc_attr_map[doc_attr_name] = model.SchemaAttributeDoc(
                name=doc_attr_name,
                type=model.Type(type_str=doc_attr_type),
                is_optional=doc_attr_optional,
                default_value=doc_attr_default,
                doc=doc_attr_desc,
            )

        # From source code: collect the attribute map from the schema stmt: get each attribute's type, default value, optional info
        code_attr_map: typing.Dict[str, model.SchemaAttributeDoc] = {}
        for attr_name, attr_obj in self.schema_type.attr_obj_map.items():
            if attr_name == objpkg.SCHEMA_SETTINGS_ATTR_NAME:
                # ignore __settings__ attribute
                # todo: show __settings__ config in schema summary info
                continue
            code_attr_type: model.Type = self.type_doc(attr_obj.attr_type)
            # fixme: hacky. the schema type str should be calculated recursively
            if (
                code_attr_type.type_category == model.Type.TypeCategory.SCHEMA
                and attr_obj.attr_node is not None
                and attr_obj.attr_node.type_node is not None
            ):
                attr_node = typing.cast(ast.SchemaAttr, attr_obj.attr_node)
                code_attr_type.type_str = attr_node.type_node.plain_type_str
            code_attr_optional: bool = attr_obj.is_optional
            # todo: get default value from ast
            code_default_value: str = ""
            code_attr_map[attr_name] = model.SchemaAttributeDoc(
                name=attr_name,
                type=code_attr_type,
                is_optional=code_attr_optional,
                default_value=code_default_value,
                doc="",
            )

        # Merge attributes from docstring and source code
        # According to source code:
        # the attribute list, the type and the optional info of each attribute
        # According to docstring:
        # the default value and the description of each attribute
        for attr_name, code_attr in code_attr_map.items():
            code_attr.doc = (
                doc_attr_map[attr_name].doc if attr_name in doc_attr_map else ""
            )
            code_attr.default_value = (
                doc_attr_map[attr_name].default_value
                if attr_name in doc_attr_map
                else ""
            )
            doc.attributes.append(code_attr)

        # Examples
        examples = ""
        for example in schema_doc["Examples"]:
            examples += example + "\n"
        doc.examples = examples
        self.doc = doc
        # Validate schema attr
        if checker:
            checker(
                model.SchemaDoc(
                    name=schema.name,
                    attributes=[code_attr_map[attr] for attr in code_attr_map],
                ),
                model.SchemaDoc(
                    name=schema.name,
                    attributes=[doc_attr_map[attr] for attr in doc_attr_map],
                ),
            ).check()

    def type_doc(self, tpe: objpkg.KCLBaseTypeObject) -> model.Type:
        def _short_schema_tpe(tpe: objpkg.KCLBaseTypeObject) -> str:
            fullname = tpe.type_str()
            parts = fullname.rsplit(".", 2)
            return f"{parts[1]}.{parts[2]}" if len(parts) > 2 else fullname

        def _get_type_str(tpe: objpkg.KCLBaseTypeObject) -> str:
            if not tpe:
                return ""
            if tpe.type_kind() in type_str_mapping:
                return type_str_mapping[tpe.type_kind()](tpe)
            else:
                return tpe.type_str()

        type_str_mapping = {
            objpkg.KCLTypeKind.StrLitKind: lambda t: f'"{t.value}"',
            objpkg.KCLTypeKind.IntLitKind: lambda t: f"{t.value}",
            objpkg.KCLTypeKind.BoolLitKind: lambda t: f"{t.value}",
            objpkg.KCLTypeKind.FloatLitKind: lambda t: f"{t.value}",
            objpkg.KCLTypeKind.NoneKind: lambda t: "None",
            objpkg.KCLTypeKind.SchemaKind: _short_schema_tpe,
            objpkg.KCLTypeKind.ListKind: lambda t: f"[{_get_type_str(t.item_type)}]",
            objpkg.KCLTypeKind.DictKind: lambda t: f"{{{_get_type_str(t.key_type)}: {_get_type_str(t.value_type)}}}"
            if t.value_type
            else f"{{{_get_type_str(t.key_type)}:}}",
            objpkg.KCLTypeKind.UnionKind: lambda t: " | ".join(
                [_get_type_str(inner_type) for inner_type in t.types]
            ),
        }

        type_mapping = {
            # ant type
            objpkg.KCLTypeKind.AnyKind: lambda t: model.Type(
                type_str=t.type_str(), type_category=model.Type.TypeCategory.ANY
            ),
            # builtin type
            objpkg.KCLTypeKind.StrKind: lambda t: model.Type(
                type_str=t.type_str(),
                type_category=model.Type.TypeCategory.BUILTIN,
                builtin_type=model.Type.BuiltinType.STRING,
            ),
            objpkg.KCLTypeKind.IntKind: lambda t: model.Type(
                type_str=t.type_str(),
                type_category=model.Type.TypeCategory.BUILTIN,
                builtin_type=model.Type.BuiltinType.INT,
            ),
            objpkg.KCLTypeKind.BoolKind: lambda t: model.Type(
                type_str=t.type_str(),
                type_category=model.Type.TypeCategory.BUILTIN,
                builtin_type=model.Type.BuiltinType.BOOL,
            ),
            objpkg.KCLTypeKind.FloatKind: lambda t: model.Type(
                type_str=t.type_str(),
                type_category=model.Type.TypeCategory.BUILTIN,
                builtin_type=model.Type.BuiltinType.FLOAT,
            ),
            # lit type
            objpkg.KCLTypeKind.StrLitKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.LIT,
                lit_type=model.LitType(string_lit=t.value),
            ),
            objpkg.KCLTypeKind.IntLitKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.LIT,
                lit_type=model.LitType(int_lit=t.value),
            ),
            objpkg.KCLTypeKind.BoolLitKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.LIT,
                lit_type=model.LitType(bool_lit=t.value),
            ),
            objpkg.KCLTypeKind.FloatLitKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.LIT,
                lit_type=model.LitType(float_lit=t.value),
            ),
            # name constant type
            objpkg.KCLTypeKind.NoneKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.NAMED_CONSTANT,
                named_constant=model.Type.NamedConstant.NONE,
            ),
            # number multiplier type
            objpkg.KCLTypeKind.NumberMultiplierKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.NUMBER_MULTIPLIER,
            ),
            # schema type
            objpkg.KCLTypeKind.SchemaKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.SCHEMA,
                schema_type=model.SchemaType(
                    name=t.name,
                    relative_path=t.filename.replace(self.root, ".", 1),
                ),
            ),
            # list type
            objpkg.KCLTypeKind.ListKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.LIST,
                list_type=model.ListType(item_type=self.type_doc(t.item_type)),
            ),
            # dict type
            objpkg.KCLTypeKind.DictKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.DICT,
                dict_type=model.DictType(
                    key_type=self.type_doc(t.key_type),
                    value_type=self.type_doc(t.value_type),
                ),
            ),
            # union type
            objpkg.KCLTypeKind.UnionKind: lambda t: model.Type(
                type_str=_get_type_str(t),
                type_category=model.Type.TypeCategory.UNION,
                union_type=model.UnionType(
                    types=[self.type_doc(inner_type) for inner_type in t.types]
                ),
            ),
        }
        if tpe.type_kind() in type_mapping:
            return type_mapping[tpe.type_kind()](tpe)
        raise TypeError



# Copyright 2021 The KCL Authors. All rights reserved.

from .format import (
    TextAdapterWalker,
    Formatter,
    kcl_ast_to_fmt_file,
    kcl_fmt_source,
    kcl_fmt_dir,
    kcl_fmt_file,
    kcl_fmt,
)

__all__ = [
    "TextAdapterWalker",
    "Formatter",
    "kcl_ast_to_fmt_file",
    "kcl_fmt_source",
    "kcl_fmt_dir",
    "kcl_fmt_file",
    "kcl_fmt",
]



class InvalidCheckerError(Exception):
    """raised when selected reporter is invalid (e.g. not found)"""

    def __init__(self, checker: str):
        self.checker_name = checker

    def __str__(self):
        return f"Args wrong, checker {self.checker_name} not found"


class EmptyReporterError(Exception):
    """raised when reporter list is empty and so should not be displayed"""

    def __str__(self):
        return "Without output reporter"


class InvalidReporterError(Exception):
    """raised when selected reporter is invalid (e.g. not found)"""

    def __init__(self, reporter: str):
        self.reporter_name = reporter

    def __str__(self):
        return f"Args wrong, reporter {self.reporter_name} not found"



#! /usr/bin/env python3

import math


def KMANGLED_ceil(*args, **kwargs):
    return math.ceil(*args, **kwargs)


def KMANGLED_factorial(*args, **kwargs):
    return math.factorial(*args, **kwargs)


def KMANGLED_floor(*args, **kwargs):
    return math.floor(*args, **kwargs)


def KMANGLED_gcd(*args, **kwargs):
    return math.gcd(*args, **kwargs)


def KMANGLED_isfinite(*args, **kwargs):
    return math.isfinite(*args, **kwargs)


def KMANGLED_isinf(*args, **kwargs):
    return math.isinf(*args, **kwargs)


def KMANGLED_isnan(*args, **kwargs):
    return math.isnan(*args, **kwargs)


def KMANGLED_modf(x):
    return list(math.modf(x))


def KMANGLED_exp(*args, **kwargs):
    return math.exp(*args, **kwargs)


def KMANGLED_expm1(*args, **kwargs):
    return math.expm1(*args, **kwargs)


def KMANGLED_log(*args, **kwargs):
    return math.log(*args, **kwargs)


def KMANGLED_log1p(*args, **kwargs):
    return math.log1p(*args, **kwargs)


def KMANGLED_log2(*args, **kwargs):
    return math.log2(*args, **kwargs)


def KMANGLED_log10(n):
    return math.log10(n)


def KMANGLED_pow(*args, **kwargs):
    return math.pow(*args, **kwargs)


def KMANGLED_sqrt(*args, **kwargs):
    return math.sqrt(*args, **kwargs)



# Copyright 2021 The KCL Authors. All rights reserved.

import re

import kclvm.kcl.error as kcl_error

DICT_TYPE_NAME = "dict"
LIST_TYPE_NAME = "list"
SCHEMA_TYPE_NAME = "Schema"
CLASS_TYPE_TMPL = "<class '{}'>"
builtin_types = ["str", "int", "float", "bool"]


def isdicttype(expectedtype):
    if isinstance(expectedtype, str) and len(expectedtype) >= 2:
        return expectedtype[0] == "{" and expectedtype[-1] == "}"
    else:
        return False


def islisttype(expectedtype):
    if isinstance(expectedtype, str) and len(expectedtype) >= 2:
        return expectedtype[0] == "[" and expectedtype[-1] == "]"
    else:
        return False


def separate_kv(expectedtype):
    stack = ""
    n = 0
    try:
        for c in expectedtype:
            if c == "[" or c == "{":
                stack += c
            elif c == "]":
                if stack[-1] != "[":
                    raise
                stack = stack[:-1]
            elif c == "}":
                if stack[-1] != "{":
                    raise
                stack = stack[:-1]
            elif c == ":":
                if len(stack) != 0:
                    raise
                return expectedtype[:n], expectedtype[n + 1 :]
            n += 1
    except Exception:
        return None, None
    return "", ""


def dereferencetype(expectedtype):
    if (
        len(expectedtype) > 1
        and (expectedtype[0] == "[" and expectedtype[-1] == "]")
        or (expectedtype[0] == "{" and expectedtype[-1] == "}")
    ):
        return expectedtype[1:-1]
    return expectedtype


def split_type_union(type_union: str):
    """
    Split the union type e.g. 'A|B|C' -> ['A', 'B', 'C'], do not split '|' in dict and list
    """
    i = 0
    s_index = 0
    stack = []
    types = []
    while i < len(type_union):
        c = type_union[i]
        if c == "|" and len(stack) == 0:
            types.append(type_union[s_index:i])
            s_index = i + 1
        # List/Dict type
        if c == "[" or c == "{":
            stack.append(c)
        # List/Dict type
        if c == "]" or c == "}":
            stack.pop()
        # String literal type
        if c == '"':
            matched = re.match(r'"(?!"").*?(?<!\\)(\\\\)*?"', type_union[i:])
            if matched:
                i += matched.span()[1] - 1
        elif c == "'":
            matched = re.match(r"'(?!'').*?(?<!\\)(\\\\)*?'", type_union[i:])
            if matched:
                i += matched.span()[1] - 1
        i += 1
    if len(stack) != 0:
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.EvaluationError_TYPE,
            arg_msg="invalid bracket matching {}".format(stack[-1]),
        )
    types.append(type_union[s_index:])
    return types


def is_type_union(tpe: str) -> bool:
    """Whether a type string is a union type string, e.g. A|B|C,
    and detect '|' in type string except '|' in dict or list.
    """
    stack = []
    i = 0
    while i < len(tpe or ""):
        c = tpe[i]
        if c == "|" and not stack:
            return True
        if c in "[{":
            stack.append(c)
        if c in "]}":
            stack.pop()
        if c == '"':
            matched = re.match(r'"(?!"").*?(?<!\\)(\\\\)*?"', tpe[i:])
            if matched:
                i += matched.span()[1] - 1
        elif c == "'":
            matched = re.match(r"'(?!'').*?(?<!\\)(\\\\)*?'", tpe[i:])
            if matched:
                i += matched.span()[1] - 1
        i += 1
    return False


def is_builtin_type(tpe_str):
    return tpe_str in builtin_types


def get_tpes_str(tpes):
    if not tpes:
        return tpes
    tpes_str = ""
    idx = 0
    for tpe in tpes:
        if idx < len(tpes) - 1:
            tpes_str += "{}, ".format(demangle_type(tpe))
        else:
            tpes_str += "{}".format(demangle_type(tpe))
        idx += 1
    return tpes_str


def get_class_name(tpe_str):
    if "class" in tpe_str:
        return re.match(r"<class \'(.*)\'>", tpe_str).group(1)
    else:
        return tpe_str


def get_builtin_type(tpe_str):
    if "kclvm_runtime.builtins" in tpe_str:
        return re.match(r"kclvm_runtime.builtins.(.*)", tpe_str).group(1)
    else:
        return tpe_str


def demangle_type(tpe: str, var=None):
    if not tpe:
        return tpe
    if isdicttype(tpe):
        key_tpe, value_tpe = separate_kv(dereferencetype(tpe))
        return "{{{}:{}}}".format(demangle_type(key_tpe), demangle_type(value_tpe))
    elif islisttype(tpe):
        return "[{}]".format(demangle_type(dereferencetype(tpe)))
    else:
        return get_builtin_type(tpe)


def do_union(obj, delta):
    """
    Union delta to obj recursively
    """
    obj_tpe = get_class_name(str(type(obj)))
    delta_tpe = get_class_name(str(type(delta)))
    if isinstance(obj, list):
        if not isinstance(delta, list):
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.TypeError_Runtime_TYPE,
                arg_msg="union failure, expect list, got {}".format(
                    demangle_type(delta_tpe, delta)
                ),
            )
        length = len(obj) if len(obj) > len(delta) else len(delta)
        result_list = obj
        for idx in range(length):
            if idx >= len(obj):
                result_list.append(delta[idx])
            elif idx < len(delta):
                result_list[idx] = union(result_list[idx], delta[idx])
        return result_list
    if isinstance(obj, dict):
        if not isinstance(delta, dict):
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.TypeError_Runtime_TYPE,
                arg_msg="union failure, expect dict, got {}".format(
                    demangle_type(delta_tpe, delta)
                ),
            )
        result_dict = obj
        for k in delta:
            if k not in obj:
                result_dict[k] = delta[k]
            else:
                result_dict[k] = union(obj[k], delta[k])
        return result_dict
    if type(obj) != type(delta):
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.TypeError_Runtime_TYPE,
            arg_msg="union failure, expect {}, got {}".format(
                demangle_type(obj_tpe), demangle_type(delta_tpe, delta)
            ),
        )
    return delta


def union(obj, delta, or_mode=False):
    if obj is None:
        return delta
    if delta is None:
        return obj
    if isinstance(obj, (list, dict)) or isinstance(delta, (list, dict)):
        return do_union(obj, delta)
    if or_mode:
        return obj | delta
    else:
        return obj if delta is None else delta



import os
import glob
import pathlib
from io import StringIO
from typing import List, Dict, Optional
from dataclasses import dataclass

import kclvm.compiler.parser.parser as parser
import kclvm.compiler.vfs as vfs
import kclvm.compiler.extension.plugin.plugin_model as plugin
import kclvm.compiler.extension.builtin.builtin as builtin
import kclvm.kcl.ast.ast as ast
import kclvm.kcl.info as kcl_info
import kclvm.tools.printer.printer as printer


@dataclass
class Config:
    name_len: int = 30
    type_len: int = 30
    default_len: int = 30
    final_len: int = 10
    optional_len: int = 10


def get_import_module(
    module: ast.Module, result: Dict[str, ast.Module] = None
) -> Optional[Dict[str, ast.Module]]:
    """Get all import module in a module."""
    if not module:
        return None
    assert isinstance(module, ast.Module)
    if not result:
        result = {}
    import_file_list = []
    import_stmt_list = module.GetImportList()
    work_dir = os.path.dirname(module.filename)
    root: str = vfs.GetPkgRoot(work_dir)
    if not root:
        root = work_dir
    for stmt in import_stmt_list or []:
        if (
            stmt.path.startswith(plugin.PLUGIN_MODULE_NAME)
            or stmt.name in builtin.STANDARD_SYSTEM_MODULES
        ):
            continue
        # import_path to abs_path
        fix_path = vfs.FixImportPath(root, module.filename, stmt.path).replace(
            ".", os.sep
        )
        abs_path = os.path.join(root, fix_path)
        # Get all .k file if path is a folder
        if os.path.isdir(abs_path):
            file_glob = os.path.join(abs_path, "**", kcl_info.KCL_FILE_PATTERN)
            import_file_list += glob.glob(file_glob, recursive=True)
        else:
            abs_path += kcl_info.KCL_FILE_SUFFIX
            import_file_list.append(abs_path)
    for file in import_file_list:
        # Skip `_*.k` and `*_test.k` kcl files
        if os.path.basename(file).startswith("_"):
            continue
        if file.endswith("_test.k"):
            continue
        if file not in result:
            import_module = parser.ParseFile(file)
            result[file] = import_module
            if import_module.GetImportList():
                get_import_module(import_module, result)
    return result


def get_import_schema(
    module: ast.Module,
) -> Optional[Dict[ast.Module, List[ast.SchemaStmt]]]:
    """Get all import schema in a module."""
    if not module:
        return None
    assert isinstance(module, ast.Module)
    import_module_list = get_import_module(module).values()
    import_schema_map = {m: m.GetSchemaList() for m in import_module_list}
    return import_schema_map


class FullSchema(ast.SchemaStmt):
    """
    Schema with base schema's attr.
    todo： mixin attr
    """

    def __init__(self, schema: ast.SchemaStmt, module: ast.Module) -> None:
        super().__init__(schema.line, schema.column)
        self.self_schema = schema
        self.parent_attr: Dict[str, List[ast.SchemaAttr]] = get_parent_attr_map(
            schema, module, {}
        )

    def __str__(self):
        s = self.self_schema.name + ", attr:["
        for name in self.self_schema.GetAttrNameList():
            s += f"{name}, "
        s = s[:-2] + "],"
        for p in self.parent_attr:
            s += f" parent:{p}, attr:["
            for attr in self.parent_attr[p]:
                s += f"{attr.name}, "
            s = s[:-2] + "],"

        return s


def get_parent_attr_map(
    ori_schema: ast.SchemaStmt,
    module: ast.Module,
    result: Dict[str, List[ast.SchemaAttr]] = None,
) -> Optional[Dict[str, List[ast.SchemaAttr]]]:
    if not ori_schema or not module:
        return None
    assert isinstance(ori_schema, ast.SchemaStmt)
    assert isinstance(module, ast.Module)
    if not result:
        result = {}
    if not ori_schema.parent_name:
        return result
    else:
        # Current module and schema.
        full_schema_map: Dict[ast.Module, List[ast.SchemaStmt]] = {
            module: module.GetSchemaList()
        }
        # Import module and schemas.
        full_schema_map.update(get_import_schema(module))
        # key : module , value: List[ast.SchemaStmt]
        for key, value in full_schema_map.items():
            for schema in value:
                if schema.name == ori_schema.parent_name.get_name():
                    result[schema.name] = schema.GetAttrList()
                    if schema.parent_name:
                        get_parent_attr_map(schema, key, result)
                    break
            else:
                continue
            break
    return result


def get_full_schema_list(module: ast.Module) -> List[FullSchema]:
    """Get all FullSchema in a module"""
    schema_list = module.GetSchemaList()
    full_schema_list = [FullSchema(schema, module) for schema in schema_list]
    return full_schema_list


class ListAttributePrinter:
    def __init__(self, file: str = None, config: Config = Config()) -> None:
        self.file = file
        self.name_len = config.name_len
        self.type_len = config.type_len
        self.default_len = config.default_len
        self.final_len = config.final_len
        self.optional_len = config.optional_len
        self.module = None
        self.schema_list = None
        self.import_schema_list = None
        self.full_schema_list = None

    def build_full_schema_list(self):
        self.module = parser.ParseFile(self.file)
        self.schema_list = self.module.GetSchemaList()
        self.import_schema_list = get_import_schema(self.module)
        self.full_schema_list = get_full_schema_list(self.module)

    def print(self):
        self.build_full_schema_list()
        if self.module:
            self.print_schema_list()
            self.print_schema_structures()

    def print_schema_list(self):
        print("------------ schema list ------------")
        file_path = self.module.filename
        file_name = pathlib.Path(file_path).name
        print("Here are schemas defined in {}:".format(file_name))
        for schema in self.schema_list:
            print("- " + schema.name)
        print("Here are schemas imported to {}:".format(file_name))
        for key, value in self.import_schema_list.items():
            import_file_path = key.filename
            import_file_name = pathlib.Path(import_file_path).name
            if len(value) > 0:
                print("imported from {}".format(import_file_name))
                for schema in value:
                    print("- " + schema.name)

    def print_schema_structures(self):
        print("------------ schema structures ------------")
        for full_schema in self.full_schema_list:
            print("schema {}:".format(full_schema.self_schema.name))
            self.print_header()
            for attr in full_schema.self_schema.GetAttrList():
                self.print_schema_attr(attr)

            for key, value in full_schema.parent_attr.items():
                print("attrs inherited from {}".format(key))
                for attr in value:
                    self.print_schema_attr(attr)
            print()

    def _print_schema_attr(self, attr: ast.SchemaAttr, default: str):
        print(
            "{:<{}}{:<{}}{:<{}}{:<{}}{:<{}}".format(
                # name
                attr.name
                if len(attr.name) <= self.name_len
                else attr.name[: self.name_len - 3] + "...",
                self.name_len,
                # type
                attr.type_str
                if len(attr.type_str) <= self.type_len
                else attr.type_str[: self.type_len - 3] + "...",
                self.type_len,
                # default
                default,
                self.default_len,
                "",
                self.final_len,
                # is_optional
                "" if attr.is_optional else "Required",
                self.optional_len,
            )
        )

    def print_schema_attr(self, attr: ast.SchemaAttr):
        if not attr:
            return
        assert isinstance(attr, ast.SchemaAttr)
        if attr.value and isinstance(attr.value, ast.SchemaExpr):
            """
            Because ast node SchemaExpr is too long to print,
            when the default value of attr.value is a SchemaExpr,just print schema name,e.g.:
            schema Name:
                firstName : str
                lastName : str

            schema Person:
                name: Name = Name {
                    firstName = "hello"
                    lastName = "world"
                }
            -------------------------------------
            schema Person:
            name                type                          default  ...
            name                Name                      ->  Name{...}
            """
            default = (
                attr.type_str
                if len(attr.type_str) <= (self.default_len - 5)
                else attr.type_str[: self.default_len - 5]
            ) + "{...}"
            self._print_schema_attr(attr, default)
            return
        with StringIO() as expr:
            printer.PrintAST(attr.value, expr)
            default_str = expr.getvalue()
            if len(default_str) > self.default_len or ("\n" in default_str):
                default_str = "..."
            self._print_schema_attr(attr, default_str)

    def print_header(self):
        print(
            "{:<{}}{:<{}}{:<{}}{:<{}}{:<{}}".format(
                # name
                "name",
                self.name_len,
                # type
                "type",
                self.type_len,
                # default
                "default",
                self.default_len,
                # is_final
                "is_final",
                self.final_len,
                # is_optional
                "is_optional",
                self.optional_len,
            )
        )



#! /usr/bin/env python3

from datetime import datetime as dt
import time as _time


def KMANGLED_today():
    """
    Return the datetime today
    """
    return str(dt.today())


def KMANGLED_now():
    """
    Return the time at now
    """
    return _time.asctime(_time.localtime(_time.time()))


def KMANGLED_ticks() -> float:
    """
    Return the current time in seconds since the Epoch.
    """
    return _time.time()


def KMANGLED_date() -> str:
    """
    Return the datetime string
    """
    return _time.strftime("%Y-%m-%d %H:%M:%S", _time.localtime())
